@@ -3,24 +3,131 @@
 import os
 import sys
 import logging
-from collections import namedtuple
+from collections import namedtuple, UserDict
 from bisect import bisect_left
 
-from langutil import C_OK, C_FAIL, C_ERROR, C_UNKNOWN
+from langutil import C_OK, C_FAIL, C_ERROR, C_UNKNOWN, C_FILE_NOT_FOUND
 
 from pypos3d.wftk.PoserFileParser import PoserFileParser
 
 from pypos3d.pftk.PoserBasic import cleanName, getRealPath, index, PoserConst, PoserToken, Lang, create, WBoolLine
-from pypos3d.pftk.SimpleAttribut import LinkParmsSA
-from pypos3d.pftk.StructuredAttribut import GenericTransform, ValueOpDelta, Deltas, ChannelMorphStatus, ChannelMorphStatusList, PoserMaterial, Keys, calcMapping_KDTree, findNewDelta, filterLength, CustomData
-from pypos3d.pftk.GeomCustom import GeomCustom
+from pypos3d.pftk.SimpleAttribut import LinkParmsSA, SimpleAttribut
+from pypos3d.pftk.StructuredAttribut import GenericTransform, ValueOpDelta, ChannelMorphStatus, ChannelMorphStatusList, PoserMaterial, Keys, calcMapping_KDTree, CustomData
+from pypos3d.pftk.GeomCustom import GeomCustom, FigureResFile
 from pypos3d.pftk.PoserMeshed import PoserProp, PoserActor, BaseProp, PoserMeshedObject
-from pypos3d.pftk.ChannelImportAnalysis import ChannelImportAnalysis
+from pypos3d.pftk.ChannelImportAnalysis import FigurePart, ChannelImportAnalysis
+from pypos3d.wftk.WaveGeom import WaveGeom, readGeom
+
+
 
 
 # Internal tuple used to store addChild and weld links between figure's parts
 PMOLink = namedtuple('PMOLink', ['son', 'parent', 'sonPMO', 'parentPMO'])
 
+class LinkDir(UserDict):
+  ''' Internal dictionary of links.
+  if hierarchical is True, parent is a single string
+  else parent contains a list of strings (for multiple welded parents)
+    and parentPMO is the reflecting list of PoserMeshed objects
+  '''
+  
+  C_LNK_CHILD, C_LNK_WELD = 'parented', 'welded'
+  
+  def __init__(self, initialdata=None, hierarchical=True):
+    super(LinkDir, self).__init__( { } if not initialdata else initialdata )
+    self.hierarchical = hierarchical
+    
+  def lnkAdd(self, sonName:str, parentName:str, sonobj, parentobj):
+    prev = self.get(sonName)
+    if self.hierarchical:
+      if prev:
+        # Child-Parent found --> Update for childs
+        logging.warning('Part[%s] already %s to [%s]', sonName, LinkDir.C_LNK_CHILD, parentName)
+
+      self[sonName] = PMOLink(sonName, parentName, sonobj, parentobj)
+    else: # weld
+      if prev:
+        prev.parent.append(parentName)
+        prev.parentPMO.append(parentobj)
+      else:
+        self[sonName] = PMOLink(sonName, [parentName,], sonobj, [parentobj, ])
+
+  def changeReference(self, oldPartName, newPartName):
+    sonName, parentName, pPMO = '', '', None
+    
+    # FIX: Python 3.8 does not support dictionnary change during iteration
+    # for k,v in dico.items():
+    for (k,v) in [ (k,v) for k,v in self.items() ]:
+      if oldPartName==v.son:
+        del self[k]
+        sonName = newPartName
+        replace = True
+      else:
+        sonName = v.son
+        replace = False
+        
+      pPMO = v.parentPMO
+      parentName = v.parent
+      
+      if self.hierarchical:
+        if oldPartName==v.parent:
+          parentName = newPartName
+          pPMO = None # Lost Link
+          replace = True
+      else:
+        for i,p in enumerate(v.parent):
+          if oldPartName==p:
+            v.parent[i] = newPartName
+            pPMO[i] = None # lost link
+            break
+
+      if replace:
+        self[sonName] = PMOLink(sonName, parentName, v.sonPMO, pPMO)
+        
+  def checkActorDico(self, fig):
+    msg = LinkDir.C_LNK_CHILD if self.hierarchical else LinkDir.C_LNK_WELD
+    nberrors = 0
+    rebuild = False
+    for k,(son, parent, sonPMO, parentPMO) in self.items():
+      if not sonPMO:
+        sonPMO = fig.findActor(son)
+        if sonPMO:
+          # Create a new link
+          rebuild = True
+        else:
+          nberrors+=1
+          logging.warn('Missing %s Child Actor [%s] for [%s]', msg, son, parent)
+
+      if self.hierarchical:
+        if not parentPMO:
+          parentPMO = fig.findActor(parent)
+          if parentPMO:
+            rebuild = True
+          else:
+            nberrors+=1
+            logging.warn('Missing %s Parent Actor [%s] for [%s]', msg, son, parent)
+      else:
+        for i,p in enumerate(parent):
+          if not parentPMO[i]:
+            parentPMO[i] = fig.findActor(p)
+        
+      if rebuild:
+        self[k] = PMOLink(son, parent, sonPMO, parentPMO)
+  
+    return nberrors
+
+
+  def writeData(self, fw, nextPfx):
+    if self.hierarchical:
+      for lnk in self.values():
+        fw.write(nextPfx+"addChild "+lnk.son+'\n\t\t\t' + lnk.parent+ '\n')
+    else: 
+      for lnk in self.values():
+        for otherName in lnk.parent:
+          fw.write(nextPfx+"weld "+lnk.son+'\n\t\t\t'+otherName+'\n')
+
+
+
 # 
 class Figure():
   ''' This class represents a Poser Figure.  '''
@@ -44,19 +151,19 @@ class Figure():
     #self._bodyIdx = idx
 
     #  figureResFile of the figure (single and shared with the PoserFile)
-    self._bodyIdx = PoserConst.C_BAD_INDEX
+    self._bodyIdx = PoserConst.C_BAD_INDEX if idx==0 else idx
     self._lstActor = [ ] # Contain all parts of a Figure, including props and deformers
     #self._lstProp = [ ] # Attribut removed in V0.2
     #self._figDesc = None# Attribut removed in V0.2, replaced by a set of Figure class attributes
 
     self._printName = ''
     self._root = None # Root Actor Name
-    self._childs = { } # Map (child's tree) sonName:PMOLink(son, parent, PoserMeshedObject)
+    self._childs = LinkDir() # Map (child's tree) sonName:PMOLink(son, parent, PoserMeshedObject)
     self._inkyChains = [ ] # List of inkyChain
     self._defaultPick = None
     self._displayOn = True
     self._linkParms = [ ]
-    self._welds  = { } # Map (welding tree) sonName:PMOLink(son, parent, PoserMeshedObject)
+    self._welds  = LinkDir(hierarchical=False) # Map (welding dependencies - not a tree -) sonName:PMOLink(son, parent, PoserMeshedObject)
     self._allowsBending = 1
     self._figureType = 1318
     self._origFigureType = 1318
@@ -90,7 +197,13 @@ class Figure():
     return [ p for p in self.getActors() if isinstance(p, PoserProp) ]
 
   def getRoot(self): return self._root
+  
+  def setRoot(self, r):
+    self._root = r
 
+  def getRootActor(self):
+    return self.findActor(self._root)
+    
   #    
   def findActor(self, actorName, withIndex=True):
     ''' Return the actor of the given name (full name with :index) 
@@ -182,8 +295,8 @@ class Figure():
     
     #  Change at figure definition level
     #  Change Attributes : addChild and weld
-    self.__changeReference(self._childs, oldPartName, newPartName)
-    self.__changeReference(self._welds, oldPartName, newPartName)
+    self._childs.changeReference(oldPartName, newPartName)
+    self._welds.changeReference(oldPartName, newPartName)
 
     for po in self._inkyChains:        
       for po2 in po.getLstAttr():
@@ -550,46 +663,22 @@ class Figure():
     int
       C_OK : Success
       C_ERROR : Error in parameters
+      C_FILE_NOTFOUND : Some WaveFront files were not found
     '''
     ret = C_OK
 
-    if isinstance(refFigObj, Figure):
+    if not setTargetMorph:
+      return C_ERROR
 
+    if isinstance(refFigObj, Figure):
+      refMeshedObj = refFigObj
       if not lstRefGeom:
-        refPoserObject = refFigObj.getPoserFile()
-        lstRefGeom = refPoserObject.extractAll(poserRootDir, refFigObj.getBodyIndex(), None)
+        pf = refFigObj.getPoserFile()
+        lstRefGeom = pf.extractAll(poserRootDir, refFigObj.getBodyIndex(), None)
   
       #  Extract and Sort GeomCustom to get single numbered groups
       curPoserObject = self.getPoserFile()
       lstCurGeom = curPoserObject.extractAll(poserRootDir, self._bodyIdx, None)
-      if (lstCurGeom == None) or (lstRefGeom == None) or (setTargetMorph == None):
-          return C_ERROR
-  
-      for srcGC in lstCurGeom:
-        srcWG = srcGC.getWaveGeom()
-  
-        tabMapping = calcMapping_KDTree(srcWG, lstRefGeom, ropt.translation, ropt.maxDist)
-  
-        #  Find the actor
-        srcActor = curPoserObject.findActor(srcGC.getName())
-        #channelsAttr = srcActor.getChannels()
-  
-        #  Foreach morph target ==> Exploit tabMapping for srcGC
-        for channelName in setTargetMorph:
-          setFoundDeltas = findNewDelta(tabMapping, refFigObj, channelName)
-          setNewDeltas = { }
-  
-          #  Local Enhancement
-          Deltas.enhancement(srcWG, setFoundDeltas, setNewDeltas, ropt.enhance, ropt.boundingType, ropt.useVicinityLoop, ropt.alpha)
-          if len(setNewDeltas) > 0:
-            #  Filter the result list to remove too short vectors
-            setFinalDeltas = filterLength(setNewDeltas, ropt.minVectLen)
-  
-            #  Create or update the channel list
-            srcActor.updateOrCreate(channelName, self.getRefName(), srcWG, setFinalDeltas, self._bodyIdx)
-            #  Check at BODY level, if valueParm exists
-            self.updateOrCreateValueParm(channelName)
-            logging.info("Deltas created for %s.%s : %d", srcGC.getName(), channelName, len(setFinalDeltas))
 
     else: # refFigObj should be a PoserMeshedObject
       refMeshedObj = refFigObj
@@ -599,39 +688,22 @@ class Figure():
       lstCurGeom = curPoserObject.extractAll(poserRootDir, self._bodyIdx, None)
       lstRefGeom = [ refMeshedObj.getBaseGeomCustom(poserRootDir), ]
 
-      if (lstCurGeom == None) or (lstRefGeom == None) or (setTargetMorph == None):
-        return C_ERROR
+    if not min( [ gc.isValid() for gc in lstCurGeom+lstRefGeom ] ):
+      return C_FILE_NOT_FOUND
 
-      for srcGC in lstCurGeom:
-        srcWG = srcGC.getWaveGeom()
+    for srcGC in lstCurGeom:
+      srcWG = srcGC.getWaveGeom()
 
-        tabMapping = calcMapping_KDTree(srcWG, lstRefGeom, ropt.translation, ropt.maxDist)
+      tabMapping = calcMapping_KDTree(srcWG, lstRefGeom, ropt.translation, ropt.maxDist)
 
-        #  Find the actor
-        #       List<PoserActor> lsta = curPoserObject.findAllActor(srcGC.getName());
-        srcActor = curPoserObject.findActor(srcGC.getName())
-        #channelsAttr = srcActor.getChannels()
+      #  Find the actor
+      srcActor = curPoserObject.findActor(srcGC.getName())
 
-        #  Foreach morph target ==> Exploit tabMapping for srcGC
-        for channelName in setTargetMorph:
-          logging.info("Creating deltas for " + srcGC.getName() + "." + channelName)
-          setFoundDeltas = findNewDelta(tabMapping, refMeshedObj, channelName)
-          setNewDeltas = { } # new ArrayList<DeltaPoint>(lstFoundDeltas.size());
-
-          #  Local Enhancement
-          Deltas.enhancement(srcWG, setFoundDeltas, setNewDeltas, ropt.enhance, ropt.boundingType, ropt.useVicinityLoop, ropt.alpha)
-          if len(setNewDeltas) > 0:
-            #  Filter the result list to remove too short vectors
-            setFinalDeltas = filterLength(setNewDeltas, ropt.minVectLen)
-
-            #  Create or update the channel list
-            srcActor.updateOrCreate(channelName, self.getRefName(), srcWG, setFinalDeltas, self._bodyIdx)
-            #  Check at BODY level, if valueParm exists
-            self.updateOrCreateValueParm(channelName)
-            logging.info("Deltas created for %s.%s : %d", srcGC.getName(), channelName, len(setFinalDeltas))
+      srcActor.findReportDeltas(refMeshedObj, srcGC, srcWG, tabMapping, setTargetMorph, ropt, checkBODYParm=True)
       
     return ret
 
+  # TODO: Rework!!
   def updateOrCreateValueParm(self, channelName):
     ''' Create or Update a valueParm channel.
     Make it visible (not Hidden)
@@ -641,12 +713,6 @@ class Figure():
     channelName : str
       Name of the channel to create or modify.
 
-    refFigure : str
-       name of the figure that contains initial deltas
-
-    setTargetMorph : set or list of str
-      Morphs' names to create
-
     Returns
     -------
     GenericTransform
@@ -701,20 +767,26 @@ class Figure():
     res = C_OK
 
     if isinstance(pa, PoserMeshedObject):
-      lstdesc = self.getDescendant(pa)
-      lstdesc.insert(0, pa)
-
+      lstdesc = [pa, ] + self.getDescendant(pa)
+      setDelName = set()
       for act in lstdesc:
         try:
           actorName = act.getName()
           self._lstActor.remove(act)
-
-          #  Clean Figure Attributs : addChild and weld
-          self._childs = { k:v for k,v in self._childs.items() if (v.son!=actorName) and (v.parent!=actorName) }
-          self._welds = { k:v for k,v in self._welds.items() if (v.son!=actorName) and (v.parent!=actorName) }
-
+          setDelName.add(actorName)
         except:
           res = C_FAIL
+          
+      # Finaly clean links (at once for the figure)
+      #  Clean Figure Attributs : addChild and weld
+      self._childs = LinkDir( { k:v for k,v in self._childs.items() if (v.son not in setDelName) and (v.parent not in setDelName) } )
+      prevDict = self._welds        
+      self._welds = LinkDir(hierarchical=False)
+      for v in [ v for v in prevDict.values() if (v.son not in setDelName)]:
+        for i,p in enumerate(v.parent): 
+          if p not in setDelName:
+            self._welds.lnkAdd(v.son, p, v.sonPMO, v.parentPMO[i])
+        
     elif isinstance(pa, PoserMaterial): # pa should be a material
       try:
         self._lstMat.remove(pa)
@@ -844,12 +916,13 @@ class Figure():
     -------
     int
     '''
-    _res = C_OK
-    if pa in self._lstActor:
-      logging.warning("Not Implemented yet")
-    else:
-      _res = PoserConst.C_ACTOR_NOTFOUND
-    return _res
+    return NotImplemented
+#     _res = C_OK
+#     if pa in self._lstActor:
+#       logging.warning("Not Implemented yet")
+#     else:
+#       _res = PoserConst.C_ACTOR_NOTFOUND
+#     return _res
 
   #
   def extractGeometry(self, poserRootDir, lstpa=None, lstch=None):
@@ -992,10 +1065,11 @@ class Figure():
     list of str
       List of unique names
     '''
-    setres = set( lnk.son for lnk in self._welds.values() if lnk.parent==firstName )
+    setres = set( lnk.son for lnk in self._welds.values() if firstName in lnk.parent )
     parentName = self.getWeldParent(firstName)
     if parentName:
-      setres.add(parentName)
+      for p in parentName: 
+        setres.add(p)
     return list(setres)
 
   # 
@@ -1080,40 +1154,69 @@ class Figure():
     cia = ChannelImportAnalysis(self, baseDir, tabChan)
     cia.checkVocab()
     cia.checkColumns()
+    
+    for cd in cia.lstChan:
+      cia.ts[0][cd.no] = cd.checkLink(self, baseDir)
+    
     return cia
 
-  # Import channel according to specifications contained by the ChannelImport Analysis.
+  
   #
-  #@param  cia          Result of file analysis
-  #@param  poserRootDir Location of Poser files
-  #
-  def importChannels(self, cia, poserRootDir):
+  def importChannels(self, lstChanDest:list, poserRootDir, hideComputed=False):
+    '''Import channel according to specifications contained by a list of ChannelDescriptor(s).
+    
+    Parameters
+    ----------    
+    lstChanDest : list
+      list of ChannelDescriptor
+      
+    poserRootDir : str
+      Location of Poser files
+      
+    hideComputed: boolean, optional, default False
+      Hide computed parameters
+    '''
     ret = C_OK
-    for ccd in cia.lstChan:
-      if cia.ts[0][ccd.nocol] == C_OK:
+    for ccd in lstChanDest:
+      if ccd.ret == C_OK:
         logging.info("Upgrading : " + ccd.act.getName())
         gt = ccd.gt
         if gt:
           logging.info("Overriding Channel[" + gt.getName() + "]")
         else:
-          gt = GenericTransform(ccd.chanType, cia.tabChan[1][ccd.nocol])
-          gt.setPrintName(gt.getName())
+          gt = GenericTransform(ccd.chanType, ccd.chanName)
+          gt.setPrintName(ccd.printName)
           ccd.act.addAttribut(gt)
           logging.info("Creating Channel[" + gt.getName() + "]")
 
         # Set standard values
-        gt.setInitValue(ccd.initValue)
+        
+        # Do not override the initValue, if it has not been given
+        if ccd.initValue!=None: 
+          gt.setInitValue(ccd.initValue)
+          gt.addKeyFrame(0, ccd.initValue)
+          
         gt.setMin(ccd.min)
         gt.setTrackingScale(ccd.trackingScale)
         gt.setForceLimits(1)
         gt.setInterpStyleLocked(1)
-        gt.addKeyFrame(0, ccd.initValue)
-
+        gt.setHidden( (hideComputed and ccd.lstOps) or ccd.isHidden )
+        
         # Create the list of Alternate Geometries
         if ccd.chanType == PoserToken.E_geomChan:
           for fic in ccd.lstAltFiles:
             ccd.act.addAltGeom(altGeomFile=fic, poserRootDir=poserRootDir)
           ccd.max = len(ccd.act.getAltGeomList())
+         
+        elif ccd.chanType== PoserToken.E_targetGeom: # Perform a Morph import
+          wg = readGeom(ccd.lstAltFiles[0].getCanonicalPath())
+          grp = wg.getGroup(ccd.groupName) if ccd.groupName else wg.getGroups()[0]
+          if not grp:
+            logging.warning("  Morph Group not found for Channel[%s] : %s %s", gt.getName(), ccd.lstAltFiles[0], ccd.groupName if ccd.groupName else 'default')
+            ret = min(ret, C_FAIL)
+            continue
+          ret = min(ret, ccd.act.createMorph(poserRootDir, grp, ccd.chanName, altGeomNo=ccd.altGeomNo)) # , minVectLen=minDist)
+          
 
         # The number of alternate geom, could have changed
         gt.setMax(ccd.max)
@@ -1122,7 +1225,7 @@ class Figure():
           gt.addVOD(vod)
 
       else:
-        logging.info("Bypassing column : " + str(ccd.nocol))
+        logging.info("Skipping channel:%s (%d)", ccd.chanName, ccd.no)
     return ret
 
   def getChannel(self, pa, expr):
@@ -1132,12 +1235,16 @@ class Figure():
    
     NB :if the optional body index is different form the body index of the figure, 
     the returned channel could belong to another figure or prop.
+    
     Parameters
     ----------
     expr : str
-      Expression to find
+      Expression defining the channel to find
        
-    Returns null if not found
+    Returns 
+    -------
+    GenericTransform : 
+      The found channel or null if not found
     '''
 
     channelName = expr
@@ -1347,7 +1454,7 @@ class Figure():
      { vod for act in ldesc for po in act.getChannels() for vod in po.getVOD()\
           if vod.getGroupName() not in ldescNames }
 
-    # Recreate chanels in the destination figure (may not be only at BODY level)
+    # Recreate channels in the destination figure (may not be only at BODY level)
     for vod in setExternChannels:
       targetActorName = cleanName(vod.getGroupName()) + idxSuffix
       act = self.findActor(targetActorName)
@@ -1376,23 +1483,13 @@ class Figure():
     return ret
 
 
-  def __lnkAdd(self, dico, msg, sonName:str, parentName:str):
-    prev = dico.get(sonName)
-    if prev:
-      # Child-Parent found --> Update
-      logging.warning('Part[%s] already %s to [%s]', sonName, msg, parentName)
-
-    son = self.findActor(sonName)
-
-    parent = self.findActor(parentName)
-
-    dico[sonName] = PMOLink(sonName, parentName, son, parent)
  
   def getWeldParent(self, partName):
+    ''' Return the list of welded parts. Never None '''
     try:
       return self._welds[partName].parent
     except KeyError:
-      return None
+      return [ ]
  
  
  
@@ -1407,67 +1504,17 @@ class Figure():
 
     if weldActors:
       for pfils in ldesc:
-        gwp = pfils.getWeldParent()
-        if gwp:
-          self.__lnkAdd(self._welds, '', pfils.getName(), gwp)
+        for gwp in pfils.getWeldParent():
+          self._welds.lnkAdd(pfils.getName(), gwp, pfils, self.findActor(gwp))
 
-  def __changeReference(self, dico, oldPartName, newPartName):
-    replace = False
-    sonName, parentName = '', ''
-    
-    # FIX: Python 3.8 does not support dictionnary change during iteration
-    # for k,v in dico.items():
-    for (k,v) in [ (k,v) for k,v in dico.items() ]:
-      if oldPartName==v.son:
-        del dico[k]
-        sonName = newPartName
-        replace = True
-      else:
-        sonName = v.son
-              
-      if oldPartName==v.parent:
-        parentName = newPartName
-        replace = True
-      else:
-        parentName = v.parent
-
-      if replace:
-        replace=False
-        dico[sonName] = PMOLink(sonName, parentName, v.sonPMO, v.parentPMO)
-        
-  def __checkActorDico(self, dico, msg):
-    nberrors = 0
-    rebuild = False
-    for k,(son, parent, sonPMO, parentPMO) in dico.items():
-      if not sonPMO:
-        sonPMO = self.findActor(son)
-        if sonPMO:
-          # Create a new link
-          rebuild = True
-        else:
-          nberrors+=1
-          logging.warn('Missing %s Child Actor [%s] for [%s]', msg, son, parent)
-
-      if not parentPMO:
-        parentPMO = self.findActor(parent)
-        if parentPMO:
-          rebuild = True
-        else:
-          nberrors+=1
-          logging.warn('Missing %s Parent Actor [%s] for [%s]', msg, son, parent)
-      
-      if rebuild:
-        dico[k] = PMOLink(son, parent, sonPMO, parentPMO)
-  
-    return nberrors
   
   def checkActorLinks(self):
     ''' Check all actors links in all figure's actors (and props) '''
     # Check the AddChild table
-    nberrors =   self.__checkActorDico(self._childs, 'hierarchical')
+    nberrors =   self._childs.checkActorDico(self)
         
     # Check the Weld table
-    nberrors += self.__checkActorDico(self._welds, 'weld')
+    nberrors += self._welds.checkActorDico(self)
  
     # Check the Root actor
     root = self.findActor(self._root)
@@ -1493,6 +1540,194 @@ class Figure():
       
     return nberrors
  
+ 
+ 
+ 
+  def createMechanical(self, lstParts:'list of FigurePart', optimize=True):
+    ''' Create the actors described by a set of FigurePart(s) into the current figure.
+    Current figure is mostly empty.
+    At the end of the process, the resulting figure's geometry is saved.
+    Individual status of FigurePart analysis are stored in their LogStatus fields.
+    
+    * 1st FigurePart shall be the figure description
+    * 2nd FigurePart shall be the BODY (aka Root Actor)
+    * Other FigurePart are either actor or just geometry to add to the previous actor
+    
+    Parameters
+    ----------
+    lstParts : list of FigurePart
+      List of Figure Parts that describe actors, fusioned geometries, materials
+      
+    lstStatus : Will be filled with a list of ret codes  
+    
+    optimize : bool, optional, default True
+      When True, the figure's main geometry is optimized before saving.
+    
+    Return
+    ------
+    int : Status of the creation
+    '''
+    
+    # Compute the max level of the figure
+    maxLevel = max( fp.level for fp in lstParts )
+    
+    figfp = lstParts.pop(0)
+    
+    frf = FigureResFile()
+    frf.setName(PoserToken.E_figureResFile.token)
+    self.setFigResFile(frf)
+    self.setFigResFileGeom(figfp.geom)
+    self.setPrintName(figfp.printName)
+    
+    curLevelActor = [None]*(maxLevel+1)
+    
+    # Create the 'Root' of the character, without any geometry
+    bodyfp = lstParts.pop(0)
+    
+    srcWG = bodyfp.geom if bodyfp.geom else WaveGeom()
+    
+    fullCopy = True
+    if bodyfp.oplst and bodyfp.oplst[0]=='add':
+      # Add the whole source geometry to the final geometry
+      # Create a Deep copy of the input geom
+      wg = srcWG.copy()
+    else:
+      # Create an empty geometry
+      wg = WaveGeom()
+      fullCopy = False
+      
+    # Set The geometry
+    frf.setGeomCustom(wg)
+    
+    idxSuffix = ":{0:d}".format(self.getBodyIndex())
+
+    bodyAct = PoserActor(Name=cleanName(bodyfp.name) + idxSuffix)
+    bodyAct.setPrintName(bodyfp.printName)
+    bodyAct.setPoserFile(self.getPoserFile())
+    self._lstActor.append(bodyAct)
+    bodyAct.setFigure(self)
+    bodyAct.setFigureResFile(self._figureResFile)
+    
+    # Create the Root actor with an empty FigurePart
+    #grp:'GeomGroup', translation=None, rotation=None, orientation=None, center=None):
+    bodyAct.createDefaultChannels(FigurePart(center='(0.0,0.0,0.0)')) # None, '', '', '', '(0.0,0.0,0.0)')
+
+    self.setRoot(bodyAct.getName())
+    self._defaultPick = bodyAct.getName()
+    
+    curLevelActor[0] = bodyAct
+    prevLevel = 0
+      
+    for fp in lstParts:
+      # logging.info('FigurePart[%d]:%s', partNo, str(fp))
+      
+      if fp.partType[:3]=='act': # Actor Creation
+        # Check level consistency
+        if fp.logCond(fp.level<=0 or curLevelActor[fp.level-1] is None, \
+                      'Unconsistent level for part:{:s}', fp.name):
+          continue
+        
+        act = PoserActor(Name=cleanName(fp.name) + idxSuffix)
+        act.setPrintName(fp.printName)
+        #  Set the new parent figure
+        act.setPoserFile(self.getPoserFile())
+        act.setFigure(self)
+        act.setFigureResFile(self._figureResFile)
+        # Manage Menu/Access to the actor
+        act.setParent(curLevelActor[fp.level-1].getName())
+        # Add (link) to parent actor in Figure
+        self._lstActor.append(act) # += ldesc
+        #  Update child and weld trees
+        self.createActorLinks([ act, ])
+        
+        curLevelActor[fp.level] = act
+        prevLevel = fp.level
+
+        act.setHidden(fp.hidden)
+        act.setAddToMenu(fp.addToMenu)
+        act.setVisible(True)
+        act.setDisplayMode('USEPARENT')
+        act.addAttribut(SimpleAttribut(PoserToken.E_backfaceCull, '0'))
+        act.addAttribut(SimpleAttribut(PoserToken.E_visibleInReflections, '1'))
+        act.addAttribut(SimpleAttribut(PoserToken.E_visibleInRender, '1'))
+        act.addAttribut(SimpleAttribut(PoserToken.E_visibleInCamera, '1'))
+        act.addAttribut(SimpleAttribut(PoserToken.E_visibleInIDL, '1'))
+        act.addAttribut(SimpleAttribut(PoserToken.E_displacementBounds, '0'))
+        act.addAttribut(SimpleAttribut(PoserToken.E_shadingRate, '0.2'))
+        act.addAttribut(SimpleAttribut(PoserToken.E_smoothPolys, '1'))
+
+        # Set the geometry
+        if fp.geom==srcWG:
+          act.setGeomType(PoserConst.GT_GLOBAL_OBJFILE)
+          act.setGeomGroupName(fp.geomGroup.getName())
+          
+          if not fullCopy: # The final geometry does not contain the GeomGroup
+            # The final geometry was empty
+            wg.addGroup(fp.geomGroup)
+            
+        else: # The geometry source for this actor is not the global source
+          if fp.oplst and 'add' in fp.oplst:
+            act.setGeomType(PoserConst.GT_GLOBAL_OBJFILE)
+            act.setGeomGroupName(fp.geomGroup.getName())
+            # Copy the GeomGroup into the final geometry
+            wg.addGroup(fp.geomGroup)
+          else:
+            # The Geometry will be loaded from another file
+            act.setGeomType(PoserConst.GT_LOCAL_OBJFILE)
+            act.setGeomFileName(fp.geom.getName())          
+
+        # Create Default Channels
+        act.createDefaultChannels(fp) # fp.geomGroup, fp.trans, fp.rot, fp.orient, fp.center))
+        
+      elif fp.partType[:3]=='geo': # Geometry add to previous actor's GeomGroup
+        # Retrieve previous actor GeomGroup in the final geometry
+        prevGrp = wg.getGroup(curLevelActor[prevLevel].getGeomGroupName()) # TODO Check LOCAL/GLOBAL if 
+        
+        # The part comes for the global source and it has been copied
+        if fp.geom==srcWG and fullCopy:
+          # The part should be in the global copied source
+          partGeom = wg.getGroup(fp.geomGroup.getName())
+          prevGrp.fusion(partGeom)
+          if 'combine' in fp.oplst:
+            # Delete the duplicate (the source geom)
+            wg.removeGroup(partGeom)
+            
+        else: # The part comes from another WaveGeom
+          prevGrp.fusion(fp.geomGroup)
+
+      elif fp.partType[:3]=='mat': # Add a Material to the figure
+        if isinstance(fp.geom, dict):
+          for wfmat in fp.geom.values():
+            self.addMaterial(PoserMaterial(wfmat))
+          
+        elif isinstance(fp.geomGroup, Figure):
+          lstMat = fp.geomGroup.getLstMaterial() if fp.name=='*' else [ fp.geomGroup.getMaterial(fp.name), ]
+          
+          for mat in lstMat: # Add each material in self
+            self.addMaterial(mat)
+            
+        else: # partGeom is supposed to be an Actor/Prop
+          if fp.geomGroup.getCustomMaterial():
+            for mat in [ a for a in fp.geomGroup.getLstAttr() \
+                         if (a.getPoserType()==PoserToken.E_material) and (fp.name=='*' or fp.name==a.getName()) ]:
+              self.addMaterial(mat)
+          else:
+            fp.status(C_FAIL, 'Actor [{:s}] has no custom material', fp.geomGroup.getName())
+      else:
+        fp.status(C_FAIL, 'Unknow part type:{:s}', fp.partType)
+    #End of for 'FigurePart'
+    
+    # Create file of the final geometry
+    ret = min(fp.ret for fp in  lstParts)
+    if ret==C_OK:
+      if optimize:
+        wg.optimizeGroups(cleaning=True)
+      
+      ret = wg.save(figfp.geom)
+    
+    return ret
+    #End of createMechanical
+ 
   # Read Figure Description Chapter
   def read(self, st):
     code,cn,val = st.getLine()
@@ -1508,8 +1743,8 @@ class Figure():
             self._root = val
 
           elif vc==PoserToken.E_addChild: # Two lines attribute
-            code2,nw = st.getFullLine()
-            self.__lnkAdd(self._childs, 'parented', val, nw)
+            _,nw = st.getFullLine()
+            self._childs.lnkAdd(val, nw, self.findActor(val), self.findActor(nw))
 
           elif vc==PoserToken.E_defaultPick:
             self._defaultPick = val
@@ -1521,8 +1756,8 @@ class Figure():
             self._linkParms.append(sa)
 
           elif vc==PoserToken.E_weld: # Two lines attribute
-            code2,nw = st.getFullLine()
-            self.__lnkAdd(self._welds, 'welded', val, nw)
+            _,nw = st.getFullLine()
+            self._welds.lnkAdd(val, nw, self.findActor(val), self.findActor(nw))
 
           elif vc==PoserToken.E_allowsBending:
             self._allowsBending = val
@@ -1560,7 +1795,7 @@ class Figure():
             self.addAttribut(sta)
 
             #  Read the opening bracket
-            code,cn,rw = st.getLine()
+            code,cn,_ = st.getLine()
             if code==PoserFileParser.TT_LEFTBACKET:
               sta.read(st)
             else:
@@ -1614,8 +1849,7 @@ class Figure():
     fw.write(nextPfx + "name " + self._printName + '\n')
     fw.write(nextPfx + "root " + self.getRoot() + '\n')
     
-    for lnk in self._childs.values():
-      fw.write(nextPfx+"addChild "+lnk.son+'\n\t\t\t' + lnk.parent+ '\n')
+    self._childs.writeData(fw, nextPfx)
 
     for po in self._inkyChains:        
       po.write(fw, nextPfx)
@@ -1626,8 +1860,7 @@ class Figure():
     for po in self._linkParms:
       po.write(fw, nextPfx)
 
-    for lnk in self._welds.values():
-      fw.write(nextPfx+"weld "+lnk.son+'\n\t\t\t'+lnk.parent+'\n')
+    self._welds.writeData(fw, nextPfx)
 
     fw.write(nextPfx + "allowsBending " + str(self._allowsBending) + '\n')
     fw.write(nextPfx + "figureType " + str(self._figureType) + '\n')
