@@ -49,12 +49,12 @@ class GeomCustom(StructuredAttribut):
       else: # src should be a filename
         self._geom = WaveGeom()
   
-        # Read full body geometry
+        # Read full geometry
         try:
           rin = getOBJFile(src)
           pfr = PoserFileParser(rin, OBJ_FNAT)
-          self._geom.readInternal(pfr);
-          self.setName(src);
+          self._geom.readInternal(pfr)
+          self.setName(src)
           rin.close()
           logging.info("File[%s] : %d lines read", src, pfr.lineno())
 
@@ -62,8 +62,7 @@ class GeomCustom(StructuredAttribut):
           self._isValid = False
           logging.warning("File[%s] does not exist (.obj/.obz)", src)
 
-    else:  # Default Empty constructor
-      #  Underlying geometry. Represented by "float" data 
+    else:  # Default Empty constructor     
       self._geom = WaveGeom()
 
   # 
@@ -96,26 +95,26 @@ class GeomCustom(StructuredAttribut):
     if self.numbTVerts != len(self._geom.texList):
       self.numbTVerts = len(self._geom.texList)
 
-  def applyDelta(self, dlt, factor, mapPt, comCoord):
+  def applyDelta(self, dlt, factor, mapPt, comCoord:'set of integer'):
     demifactor = factor / 2.0
 
     cl = self._geom.coordList
     lcl = len(cl)
-    
-    for dp in dlt.deltaSet.values():
-      noPt = mapPt[dp.noPt] if mapPt else dp.noPt 
-      if noPt < lcl:
-        if not comCoord:
+    # Performance enhancement (test comCoord) once
+    if comCoord:
+      for dp in dlt.deltaSet.values():
+        noPt = mapPt[dp.noPt] if mapPt else dp.noPt 
+        if noPt < lcl:
+          cl[noPt].scaleAdd(demifactor if noPt in comCoord else factor, dp, cl[noPt])
+        else:
+          logging.warning("Pt:%d", noPt)
+    else:
+      for dp in dlt.deltaSet.values():
+        noPt = mapPt[dp.noPt] if mapPt else dp.noPt 
+        if noPt < lcl:
           cl[noPt].scaleAdd(factor, dp, cl[noPt])
         else:
-          try:
-            comCoord.index(noPt)
-            cl[noPt].scaleAdd(demifactor, dp, cl[noPt])
-          except:
-            cl[noPt].scaleAdd(factor, dp, cl[noPt])
-
-      else:
-        logging.warning("Pt:%d", noPt)
+          logging.warning("Pt:%d", noPt)
 
 
   # @param bodyIdx
@@ -124,12 +123,11 @@ class GeomCustom(StructuredAttribut):
   # @param mapPt
   # @param comCoord
   # @param setTargetMorph    If null consider all "PBM*" else consider the targets contained in this set.
-  def findApplyDeltaInternal(self, bodyIdx, groupName, refPoserObject, mapPt, comCoord, setTargetMorph):
+  def findApplyDeltaInternal(self, bodyIdx, groupName, refPoserObject, mapPt, comCoord:'set of integer', setTargetMorph):
     # Position of ':' change when bodyIdx >= 10
     searchName = groupName if (groupName.rfind(':') == len(groupName) - (3 if bodyIdx > 9 else 2)) else groupName + ":" + bodyIdx;
 
     # 20080406 : Duplicate Channel application in V4 (caused by getChannels())
-    # for (PoserActor po : lstact)
     po = refPoserObject.findActor(searchName)
  
     for attr in po.getChannels():
@@ -143,8 +141,6 @@ class GeomCustom(StructuredAttribut):
           logging.info("Herited Factor[%s" + "] : %s f=%g", attr.getName(), vod.getGroupName(), f)
 
           # Take into account the Control Ratio (deltaAddDelta)
-          #cr = vod.getControlRatio()
-          #finalFactor += f * cr
           finalFactor = vod.calc(finalFactor, f)
           
         dltAttr = attr.getDeltas()
@@ -171,22 +167,35 @@ class GeomCustom(StructuredAttribut):
     if isinstance(PropOrBodyIdx, int):
       bodyIdx = PropOrBodyIdx
 
-      # Calculate global comCoord table    
-      comCoord = [ ]
+      # Calculate global comCoord set (better for speed)    
+      comCoord = set()
 
       # Get all part of the character
       lstActor = refPoserObject.getDescendant(bodyIdx)
-      nbtotComCoord = 0
 
       for gn in [ a.getName() for a in lstActor if a!=lstActor[0] ]:
         lstDesc = refPoserObject.getWelded(bodyIdx, gn)
-        if lstDesc:
-          tmpComCoord = self._geom.findJonction(gn, lstDesc[0])
-          if tmpComCoord:
-            nbtotComCoord += len(tmpComCoord)
-            comCoord += tmpComCoord
+        gBas = self._geom.getGroup(gn)
+        if lstDesc and gBas:
+          sBas = set(gBas.vertIdx)
+          logging.info("Groupe Bas[%s]: %d faces", gn, gBas.getNbFace())
+          # FIX 20210327: bad detection of common points --> New algo based on set()
+          for d in lstDesc:
+            # Retrieve the GeomGroup of the actor
+            gHaut = self._geom.getGroup(d)
+            if gHaut:
+              logging.info("Groupe Haut[%s]: %d faces", d, gHaut.getNbFace()) 
 
-      logging.info("Nb Common Coordinates =%d", nbtotComCoord)
+              # Common points belong to both groups
+              tmpComCoord = sBas & set(gHaut.vertIdx)
+
+              # Merge the set of interface points
+              comCoord |= tmpComCoord
+            else:
+              logging.warning("Groupe Bas[%s] NO faces", d )
+
+
+      logging.info("Applying Deltas with [%d] Common Coordinates", len(comCoord))
 
       # log.info("------------------------ Applying Deltas ----------------")
       for gn in  [ a.getName() for a in lstActor ]:
@@ -232,8 +241,8 @@ class GeomCustom(StructuredAttribut):
   def calcPointMapping(self, newGeom):
     self._geom.calcPointMapping(newGeom._geom)
 
-  def compareTo(self, gc):
-    return self._geom.compareTo(gc.getWaveGeom())
+#   def compareTo(self, gc):
+#     return self._geom.compareTo(gc.getWaveGeom())
 
   # Extract the group of the given name and create a new GeomCustom
   # that contains a <u>deep copy</u> of the original data.
@@ -308,6 +317,13 @@ class FigureResFile(SimpleAttribut):
 
     return self._geomCustom
 
+  def setGeomCustom(self, gc:'GeomCustom or WaveGeom'):
+    if isinstance(gc, WaveGeom):
+      wg = gc
+      gc = GeomCustom(src=wg)
+      
+    self._geomCustom = gc
+
   def getPath(self, poserRootDir):
     return getRealPath(poserRootDir, self.getValue())
 
