@@ -8,6 +8,7 @@ Created on 1 mai 2020
 '''
 import sys, traceback
 import os.path
+import copy
 from collections import namedtuple
 import platform
 import importlib
@@ -69,14 +70,10 @@ def getRealPath(poserRootDir, poserPath):
 
     #  Try to find the last part of poserRootDir in the path
     parts = poserRootDir.split("/")
-    lastPart = poserRootDir
-    if len(parts)>=2:
-      lastPart = parts[len(parts)-1]
-    else:
-      lastPart = parts[1]
+    lastPart = parts[len(parts)-1] if len(parts)>=2 else parts[1]
 
     lpi = poserPath.find(lastPart)
-    return poserRootDir + poserPath[lpi + len(lastPart):].replace('\\', os.sep) if (lpi >= 0) else None
+    return poserRootDir + poserPath[lpi + len(lastPart):].replace('\\', os.sep) if (lpi >= 0) else poserPath
 
   #  Avoid *nix absolute path
   if poserPath[0]==os.sep:
@@ -89,8 +86,10 @@ def getRealPath(poserRootDir, poserPath):
   
   if poserPath[0]==os.sep:
     poserPath = poserPath[1:]
+  elif len(poserPath)>2 and poserPath[0]=='.' and poserPath[1]==os.sep:
+    poserPath = poserPath[2:]
     
-  return os.path.join(poserRootDir, poserPath)
+  return poserPath if os.path.isfile(poserPath) else os.path.join(poserRootDir, poserPath)
  
 
 # Return the index of a name, i.e. the leading ":n"
@@ -318,8 +317,8 @@ class PoserObject(object):
   
   def toString(self): return self._poserType.token
         
-PToken = namedtuple('PToken', ['token', 'isStructured', 'clazz', 'isDirect'])
-PToken.__new__.__defaults__ = ('',False,None,True) 
+PToken = namedtuple('PToken', ['token', 'isStructured', 'clazz', 'isDirect', 'defaultName'])
+PToken.__new__.__defaults__ = ('',False,None,True,'') 
 
 class PoserToken:
   ''' Poser class to manage Poser File language (token) and implementation modules. ''' 
@@ -350,7 +349,7 @@ class PoserToken:
   E_shaderTree = PToken("shaderTree", True, 'ShaderTree', False) # v 5
   E_node = PToken("node", True, 'Node', False) # v 5
   E_pos = PToken("pos", False, None, False) # v5
-  E_showPreview = PToken("showPreview", False, None, False) # v5
+  E_showPreview = PToken("showPreview", False, None, True) # v5
   E_nodeInput = PToken("nodeInput", True, 'NodeInput', False) # v 5
   E_value = PToken("value", False) # v5
   E_parmR = PToken("parmR", False) # v5
@@ -523,7 +522,7 @@ class PoserToken:
   E_giSamples = PToken("giSamples", False, None, False)
   E_giDepth = PToken("giDepth", False, None, False)
   E_renderOptionsIndex = PToken("renderOptionsIndex", False, None, False)
-  E_inputsCollapsed = PToken("inputsCollapsed", False, None, False)
+  E_inputsCollapsed = PToken("inputsCollapsed", False, None, True)
   
   E_morphInfo = PToken("morphInfo", True, 'StructuredAttribut', False)
   E_morphPutty = PToken("morphPutty", True, 'StructuredAttribut', False)
@@ -581,22 +580,22 @@ class PoserToken:
   E_liteAttenStart = PToken("liteAttenStart", True, 'GenericTransform', False)
   E_liteAttenEnd = PToken("liteAttenEnd", True, 'GenericTransform', False)
   
-  E_rotateX = PToken("rotateX", True, 'GenericTransform', False)
-  E_rotateY = PToken("rotateY", True, 'GenericTransform', False)
-  E_rotateZ = PToken("rotateZ", True, 'GenericTransform', False)
+  E_rotateX = PToken("rotateX", True, 'GenericTransform', False, 'xrot')
+  E_rotateY = PToken("rotateY", True, 'GenericTransform', False, 'yrot')
+  E_rotateZ = PToken("rotateZ", True, 'GenericTransform', False, 'zrot')
   E_keys = PToken("keys", True, 'Keys', False)
   E_kdIntensity = PToken("kdIntensity", True, 'GenericTransform', False)
-  E_propagatingScale = PToken("propagatingScale", True, 'GenericTransform', False)
-  E_propagatingScaleX = PToken("propagatingScaleX", True, 'GenericTransform', False)
-  E_propagatingScaleY = PToken("propagatingScaleY", True, 'GenericTransform', False)
-  E_propagatingScaleZ = PToken("propagatingScaleZ", True, 'GenericTransform', False)
-  E_translateX = PToken("translateX", True, 'GenericTransform', False)
-  E_translateY = PToken("translateY", True, 'GenericTransform', False)
-  E_translateZ = PToken("translateZ", True, 'GenericTransform', False)
-  E_scale = PToken("scale", True, 'GenericTransform', False)
-  E_scaleX = PToken("scaleX", True, 'Scale', False)
-  E_scaleY = PToken("scaleY", True, 'Scale', False)
-  E_scaleZ = PToken("scaleZ", True, 'Scale', False)
+  E_propagatingScale = PToken("propagatingScale", True, 'GenericTransform', False, 'scale')
+  E_propagatingScaleX = PToken("propagatingScaleX", True, 'GenericTransform', False, 'xScale')
+  E_propagatingScaleY = PToken("propagatingScaleY", True, 'GenericTransform', False, 'yScale')
+  E_propagatingScaleZ = PToken("propagatingScaleZ", True, 'GenericTransform', False, 'zScale')
+  E_translateX = PToken("translateX", True, 'GenericTransform', False, 'xtran')
+  E_translateY = PToken("translateY", True, 'GenericTransform', False, 'ytran')
+  E_translateZ = PToken("translateZ", True, 'GenericTransform', False, 'ztran')
+  E_scale = PToken("scale", True, 'GenericTransform', False, 'Scale')
+  E_scaleX = PToken("scaleX", True, 'Scale', False, 'xScale')
+  E_scaleY = PToken("scaleY", True, 'Scale', False, 'yScale')
+  E_scaleZ = PToken("scaleZ", True, 'Scale', False, 'zScale')
   E_visibility = PToken("visibility", True, 'GenericTransform', False)
   
   E_kdRed = PToken("kdRed", True, 'GenericTransform', False)
@@ -630,12 +629,12 @@ class PoserToken:
   E_camAutoCenterZ = PToken("camAutoCenterZ", True, 'GenericTransform', False)
   
   E_valueParm = PToken("valueParm", True, 'GenericTransform', False)
-  E_xOffsetA = PToken("xOffsetA", True, 'GenericTransform', False)
-  E_yOffsetA = PToken("yOffsetA", True, 'GenericTransform', False)
-  E_zOffsetA = PToken("zOffsetA", True, 'GenericTransform', False)
-  E_xOffsetB = PToken("xOffsetB", True, 'GenericTransform', False)
-  E_yOffsetB = PToken("yOffsetB", True, 'GenericTransform', False)
-  E_zOffsetB = PToken("zOffsetB", True, 'GenericTransform', False)
+  E_xOffsetA = PToken("xOffsetA", True, 'GenericTransform', False, 'xOffsetA')
+  E_yOffsetA = PToken("yOffsetA", True, 'GenericTransform', False, 'yOffsetA')
+  E_zOffsetA = PToken("zOffsetA", True, 'GenericTransform', False, 'zOffsetA')
+  E_xOffsetB = PToken("xOffsetB", True, 'GenericTransform', False, 'xTranB')
+  E_yOffsetB = PToken("yOffsetB", True, 'GenericTransform', False, 'yTranB')
+  E_zOffsetB = PToken("zOffsetB", True, 'GenericTransform', False, 'zTranB')
   
   E_taperX = PToken("taperX", True, 'GenericTransform', False)
   E_taperY = PToken("taperY", True, 'GenericTransform', False)
@@ -826,14 +825,14 @@ class PoserToken:
   E_forceLimits = PToken("forceLimits", False)
   E_min = PToken("min", False)
   E_max = PToken("max", False)
-  E_tMin = PToken("tMin", False, None, False)
-  E_tMax = PToken("tMax", False, None, False)
-  E_tExpo = PToken("tExpo", False, None, False)
-  E_ksIgnoreTexture = PToken("ksIgnoreTexture", False, None, False)
-  E_reflectThruLights = PToken("reflectThruLights", False, None, False)
-  E_reflectThruKd = PToken("reflectThruKd", False, None, False)
-  E_bumpStrength = PToken("bumpStrength", False, None, False)
-  E_NsExponent = PToken("NsExponent", False, None, False)
+  E_tMin = PToken("tMin", False, None, True)
+  E_tMax = PToken("tMax", False, None, True)
+  E_tExpo = PToken("tExpo", False, None, True)
+  E_ksIgnoreTexture = PToken("ksIgnoreTexture", False, None, True)
+  E_reflectThruLights = PToken("reflectThruLights", False, None, True)
+  E_reflectThruKd = PToken("reflectThruKd", False, None, True)
+  E_bumpStrength = PToken("bumpStrength", False, None, True)
+  E_NsExponent = PToken("NsExponent", False, None, True)
   E_static = PToken("static", False, None, False)
   E_lightType = PToken("lightType", False, None, False)
   E_k = PToken("k", False, 'KSA', False)
@@ -898,7 +897,7 @@ class PoserToken:
   E_bumpMap = PToken("bumpMap", False, 'TextureMapSA', False)
   E_reflectionMap = PToken("reflectionMap", False, 'TextureMapSA', False)
   E_transparencyMap = PToken("transparencyMap", False, 'TextureMapSA', False)
-  E_reflectionStrength = PToken("reflectionStrength", False, None, False)
+  E_reflectionStrength = PToken("reflectionStrength", False, None, True)
   
   E_allowsBending = PToken("allowsBending", False, None, False)
   E_figureType = PToken("figureType", False, None, False)
@@ -936,11 +935,11 @@ class PoserToken:
   E_off = PToken("off", False)
   E_sm = PToken("sm", False, 'OffSA', False)
   E_calcWeights = PToken("calcWeights", False, 'OffSA', False)
-  E_KdColor = PToken("KdColor", False, None, False)
-  E_KaColor = PToken("KaColor", False, None, False)
-  E_KsColor = PToken("KsColor", False, None, False)
-  E_ReflectionColor = PToken("ReflectionColor", False, None, False)
-  E_TextureColor = PToken("TextureColor", False, None, False)
+  E_KdColor = PToken("KdColor", False, None, True)
+  E_KaColor = PToken("KaColor", False, None, True)
+  E_KsColor = PToken("KsColor", False, None, True)
+  E_ReflectionColor = PToken("ReflectionColor", False, None, True)
+  E_TextureColor = PToken("TextureColor", False, None, True)
   
   E_port = PToken("port", False, None, False) # Use a Quad attribut
   E_d = PToken("d", False, None, False)
@@ -983,15 +982,115 @@ class PoserToken:
   E_Zones = PToken("zones", True, "Zones", False) # Poser9
   E_WeightMapZone = PToken("weightmapzone", True, "WeightMapZone", False) # Poser9
   E_sphereZone = PToken("spherezone", True, 'SphereZone', False)
+  E_capsuleZone = PToken("capsulezone", True, 'CapsuleZone', False)
   E_active = PToken("active", False, None, False)
   E_math = PToken("math", False, None, False)
   E_mapname = PToken("mapname", False, None, False)
+  E_outCapRadR = PToken("outCapRadR", False, None, False)
+  E_outCapRadL = PToken("outCapRadL", False, None, False)
+  E_outCylRadR = PToken("outCylRadR", False, None, False)
+  E_outCylRadL = PToken("outCylRadL", False, None, False)
+  E_inCapRadR = PToken("inCapRadR", False, None, False)
+  E_inCapRadL = PToken("inCapRadL", False, None, False)
+  E_inCylRadR = PToken("inCylRadR", False, None, False)
+  E_inCylRadL = PToken("inCylRadL", False, None, False)
+  E_outLenR = PToken("outLenR", False, None, False)
+  E_outLenL = PToken("outLenL", False, None, False)
+  E_inLenR = PToken("inLenR", False, None, False)
+  E_inLenL = PToken("inLenL", False, None, False)
+  E_matrices = PToken("matrices", False, 'SphereMathsRaw', False)
   
+  # Channel initialization:
+  #    PoserType, isStatic, min, max, initValue, order XYZ, ZXY, YZX, TrackingScale
+  MINIMAL_CHANNELS = [ \
+      # The Origin of the actor (i.e. Barycentre)
+      [E_xOffsetA,   True,  -10.0, 10.0, 0.0,   0,   1,   2, 0.1], \
+      [E_yOffsetA,   True,  -10.0, 10.0, 0.0,   1,   2,   0, 0.1], \
+      [E_zOffsetA,   True,  -10.0, 10.0, 0.0,   2,   0,   1, 0.1], \
+
+      [E_scale,      False,  0.1, 100.0, 1.0,   3,   3,   3, 0.01], \
+      [E_scaleX,     False,  0.1, 100.0, 1.0,   4,   5,   6, 0.01], \
+      [E_scaleY,     False,  0.1, 100.0, 1.0,   5,   6,   4, 0.01], \
+      [E_scaleZ,     False,  0.1, 100.0, 1.0,   6,   4,   5, 0.01], \
+      
+      [E_rotateX,    False, -180.0,  180.0, 0.0,   7,   8,   9, 1.0], \
+      [E_rotateY,    False, -180.0,  180.0, 0.0,   8,   9,   7, 1.0], \
+      [E_rotateZ,    False, -180.0,  180.0, 0.0,   9,   7,   8, 1.0], \
+
+      # The oppositte vector to Barycentre
+      [E_xOffsetB,   True,  -10.0, 10.0, 0.0,  10,  11,  12, 0.1], \
+      [E_yOffsetB,   True,  -10.0, 10.0, 0.0,  11,  12,  10, 0.1], \
+      [E_zOffsetB,   True,  -10.0, 10.0, 0.0,  12,  10,  11, 0.1], \
+      
+      [E_translateX, False, -10.0, 10.0, 0.0,  13,  14,  15, 0.01], \
+      [E_translateY, False, -10.0, 10.0, 0.0,  14,  15,  13, 0.01], \
+      [E_translateZ, False, -10.0, 10.0, 0.0,  15,  13,  14, 0.01], \
+      ]
+
+  # Channel initialization:
+  #    PoserType, isStatic, min, max, initValue,       order XYZ, ZXY, YZX
+  MINIMAL_BODY_CHANNELS = [ \
+      # The Origin of the actor (i.e. Barycentre)
+      [E_xOffsetA,   True,  -1000.0, 1000.0,  0.0,   0,   1,   2, 0.1], \
+      [E_yOffsetA,   True,  -1000.0, 1000.0, 0.74,   1,   2,   0, 0.1], \
+      [E_zOffsetA,   True,  -1000.0, 1000.0,  0.0,   2,   0,   1, 0.1], \
+
+      [E_propagatingScale,      False,     0.1, 1000.0, 1.0,   3,   3,   3, 0.01], \
+      [E_propagatingScaleX,     False,     0.1, 1000.0, 1.0,   4,   5,   6, 0.01], \
+      [E_propagatingScaleY,     False,     0.1, 1000.0, 1.0,   5,   6,   4, 0.01], \
+      [E_propagatingScaleZ,     False,     0.1, 1000.0, 1.0,   6,   4,   5, 0.01], \
+      
+      [E_rotateX,    False,  -180.0,  180.0, 0.0,   7,   8,   9, 1.0], \
+      [E_rotateY,    False,  -180.0,  180.0, 0.0,   8,   9,   7, 1.0], \
+      [E_rotateZ,    False,  -180.0,  180.0, 0.0,   9,   7,   8, 1.0], \
+
+      # The oppositte vector to Barycentre
+      [E_xOffsetB,   True,  -1000.0, 1000.0, 0.0,  10,  11,  12, 0.1], \
+      [E_yOffsetB,   True,  -1000.0, 1000.0, 0.0,  11,  12,  10, 0.1], \
+      [E_zOffsetB,   True,  -1000.0, 1000.0, 0.0,  12,  10,  11, 0.1], \
+      
+      [E_translateX, False, -1000.0, 1000.0, 0.0,  13,  14,  15, 0.01], \
+      [E_translateY, False, -1000.0, 1000.0, 0.0,  14,  15,  13, 0.01], \
+      [E_translateZ, False, -1000.0, 1000.0, 0.0,  15,  13,  14, 0.01], \
+      ]
+  
+  ORENTATIONS = { 'xyz':5, 'yzx':7, 'zxy':6 }
+  
+  SCALES  = (E_scale, E_scaleX, E_scaleY, E_scaleZ,\
+            E_propagatingScale, E_propagatingScaleX, E_propagatingScaleY, E_propagatingScaleZ)
+  
+  ROT     = (E_rotateX, E_rotateY, E_rotateZ)
+  TRANS   = (E_translateX, E_translateY, E_translateZ)
+  OFFSETA = (E_xOffsetA, E_yOffsetA, E_zOffsetA)
+  OFFSETB = (E_xOffsetB, E_yOffsetB, E_zOffsetB)
+
+  @classmethod
+  def setInitVect(clz, lst, props, v):
+    ch = next(t for t in lst if t[0]==props[0])
+    ch[4] = v.x
+    ch = next(t for t in lst if t[0]==props[1])
+    ch[4] = v.y
+    ch = next(t for t in lst if t[0]==props[2])
+    ch[4] = v.z
+
+  @classmethod
+  def getMinimalChannels(cls, axisOrder='xyz', isBody=False):
+    '''
+    Return the list of ordered channels to create either for an actor or 
+    for the Root actor (usually BODY)
+    '''
+    # Get a copy of the default channels (we will modify the content)
+    lstChan = copy.deepcopy(PoserToken.MINIMAL_BODY_CHANNELS if isBody else PoserToken.MINIMAL_CHANNELS )
+    # Sort the Channel table according to actor orientation
+    col = PoserToken.ORENTATIONS[axisOrder]
+    lstChan.sort(key=lambda t: t[col])
+    return lstChan
 
 
 Lang = { getattr(PoserToken, ex).token : getattr(PoserToken, ex) for ex in dir(PoserToken) if ex[0:2]=='E_' }
 LangList = [ getattr(PoserToken, ex) for ex in dir(PoserToken) if ex[0:2]=='E_' ]
 LowerLangName = { getattr(PoserToken, ex):getattr(PoserToken, ex).token.lower()  for ex in dir(PoserToken) if ex[0:2]=='E_' }
+LowerChanTypes = { getattr(PoserToken, ex).token.lower():getattr(PoserToken, ex)  for ex in dir(PoserToken) if ex[0:2]=='E_' and getattr(PoserToken, ex).clazz=='GenericTransform' }
 
 # In Python: Implemented by a module level function in the highest 
 # Module of this object model
