@@ -552,6 +552,30 @@ class PoserFile(object):
       ret = C_ERROR
     return ret
 
+
+  def save(self, fn):
+    ''' Write a PoserFile in text mode or in compressed mode.
+    Use the extension to find the right mode.
+
+    Parameters
+    ----------
+    fn : str
+      Full path name
+
+    Returns
+    -------
+    int
+      C_OK write without error, C_ERROR a write error has occurred
+    '''
+    if isCompressed(fn):
+      ret = self.writeZ(fn)
+    else:
+      ret = self.writeFile(fn)
+
+    return ret
+
+
+
   # 
   # @see deyme.v3d.poser.PoserObject#write(java.io.FileWriter)
   #    
@@ -851,7 +875,7 @@ class PoserFile(object):
   #          Channels to take into account. If null, no delta are applied.
   # @return The list of geometries that compose the figure.
   def extractAll(self, poserRootDir, bodyIdx, channelSet):
-    lstgeom = None
+    lstgeom = [ ]
     eg = None
 
     # Read OBJ filename
@@ -860,9 +884,7 @@ class PoserFile(object):
       if channelSet:
         body.findApplyDelta(bodyIdx, self, channelSet)
 
-      lstactors = self.getDescendant(bodyIdx)
-      lstgeom = [ ]
-      for pa in lstactors:
+      for pa in self.getDescendant(bodyIdx):
         # Check if the name of the actor vs the name of the meshed part
         if pa.getGeomType()==PoserConst.GT_GLOBAL_OBJFILE:
           # geomHandlerGeom 13 hip
@@ -873,7 +895,7 @@ class PoserFile(object):
           eg = GeomCustom(getRealPath(poserRootDir, pa.getGeomFileName()))
 
         #elif pa.getGeomType()==PoserConst.GT_NONE:
-        #elif pa.getGeomType()==PoserConst.GT_INTERNAL:
+        #elif pa.getGeomType()==PoserConst.GT_INTERNAL: 
         else:
           eg = None
 
@@ -1188,7 +1210,21 @@ class PoserFile(object):
     return C_OK if nberrors==0 else C_FAIL
       
       
-      
+  def createFigure(self, name=None):
+    ''' Create a new Figure object, initialize the correct index. '''
+    bodyIdx = 1 + (max( f.getBodyIndex() for f in self.getLstFigure() ) if self.getLstFigure() else 0)
+     
+    fig = Figure(self, idx=bodyIdx)
+    
+    if name:
+      fig.setPrintName(name)
+    
+    self._lstFigure.append(fig)
+    
+    # At this stage the Figure is not consistent (no geom, no actor, ...)
+    
+    logging.info("Figure Created: [%s]:%d", name if name else '', bodyIdx)
+    return fig
       
       
       
