@@ -2,14 +2,16 @@
 # package: pftk.poser.kern
 from collections import namedtuple
 import logging
+import math
 
-from langutil import RemoveExt, C_FAIL, C_OK, C_ERROR
-from pypos3d.wftk.WFBasic import C_BAD_DELTAINDEX, C_BAD_DELTACOUNT, C_NODELTA, Vector3d
+from langutil import RemoveExt, C_FAIL, C_OK, C_ERROR, C_FILE_NOT_FOUND
+from pypos3d.wftk.WFBasic import C_BAD_DELTAINDEX, C_BAD_DELTACOUNT, C_NODELTA, Vector3d, Point3d,\
+  FEPSILON
 from pypos3d.wftk.PoserFileParser import PoserFileParser, ParsingErrorException
 from pypos3d.pftk.PoserBasic import PoserConst, PoserToken, PToken, buildRelPath, getRealPath, index, TAS, WBoolLine, create, Lang
 from pypos3d.pftk.GeomCustom import GeomCustom
 from pypos3d.pftk.StructuredAttribut import StructuredAttribut, AlternateGeom, Deltas, ValueOpDelta, ChannelMorphStatus, \
-    GenericTransform, ChannelMorphStatusList, findNewDelta, filterLength, calcMapping_KDTree, CustomData
+    GenericTransform, ChannelMorphStatusList, findNewDelta, filterLength, calcMapping_KDTree, CustomData, DeltaPoint
 
 
 #
@@ -292,6 +294,148 @@ class PoserMeshedObject(StructuredAttribut):
     vp.setHidden(isHidden)
     return vp
 
+  
+
+  def createDefaultChannels(self, fp:'FigurePart'): # grp:'GeomGroup', translation=None, rotation=None, orientation=None, center=None):
+    ''' Create the 16 minimal channels for this actor and initialize them according to the 4 textual parameters.
+    Textual parameters are supposed to be 'stripped'
+    
+    Parameters
+    ----------
+    fp: FigurePart
+      Description of the actor.
+      Used fields are: geomGroup, translation, rotation, orientation, center
+    
+    Return
+    ------
+    int : Return Code
+    
+    '''
+    ret = C_OK
+    
+    AxisOrder, tor = 'xyz', ''
+    orientation = fp.orientation
+    
+    if orientation:
+      pv = orientation.find(',')
+      AxisOrder = orientation[:pv].lower().strip() if pv>2 else orientation.lower().strip()
+
+      if AxisOrder in PoserToken.ORENTATIONS:
+        tor = orientation[pv+1:].strip() if pv>1 and pv<len(orientation) else ''
+      else:
+        AxisOrder = 'xyz'
+        tor = orientation.strip()
+      
+    # Get a copy of the default channels (we will modify the content)
+    lstChan = PoserToken.getMinimalChannels(AxisOrder, not fp.geomGroup)
+    try:
+      # Compute Actor's center
+      bc = Point3d()
+      center= fp.center
+      if center:
+        if center[0]=='(' and center[-1]==')': # Center is given by a Tuple of float
+          bc = Point3d(eval(center))
+        elif center.lower().startswith('face='): # Can be Face=Material name
+          centerFaceVx = fp.geomGroup.getFaceVertexBy(center[5:])
+          bc = Point3d.barycentre(centerFaceVx)
+        else:
+          return fp.status(C_FAIL,'Center definition error for [{:s}]:{:s}', self.getName(), center) 
+      else:
+        bc = fp.geomGroup.calcBarycentre()
+        
+      # Set the Origin and Copy it in the static channels xOffset...
+      self.setOrigin(bc)
+      PoserToken.setInitVect(lstChan, PoserToken.OFFSETA, bc)
+      PoserToken.setInitVect(lstChan, PoserToken.OFFSETB, Point3d(bc).neg())
+  
+      # Compute Actor's orientation
+      rotx, roty, rotz = 0.0,0.0,0.0
+      if len(tor)>1:
+        
+        if tor[0]=='(' and tor[-1]==')': # The orientation is given by a Tuple of float
+          vr = Point3d(eval(tor))
+          rotx, roty, rotz = vr.x, vr.y, vr.z
+          
+        elif tor.lower().startswith('face='): # Can be Face=Material name
+          vxtab = fp.geomGroup.getFaceVertexBy(tor[5:])
+          n = Point3d.triangle_normal(vxtab[0], vxtab[1], vxtab[2])
+          if AxisOrder=='yzx':
+            rotz = math.asin(-n.x)
+            cz   = math.cos(rotz)
+            rotx = math.acos(n.y / cz) if math.fabs(cz)>FEPSILON else 0.0 #
+          elif AxisOrder=='xyz':
+            roty = math.asin(n.z)
+            cy   = math.cos(roty)
+            rotz = math.acos(n.x/cy) if math.fabs(cy)>FEPSILON else 0.0
+          else: # 'zxy'
+            rotx = math.asin(-n.y)
+            cx   = math.cos(rotx)
+            roty = math.acos(n.z/cx) if math.fabs(cx)>FEPSILON else 0.0
+          
+          # Convert Result un degree
+          rotx = math.degrees(rotx)
+          roty = math.degrees(roty)
+          rotz = math.degrees(rotz)
+        else:
+          return fp.status(C_FAIL, 'Orientation definition error for [{:s}]:{:s}', self.getName(), tor)
+        
+      self.setOrientation(Vector3d(rotx, roty, rotz))
+      
+      # Manage Translation
+      tr = Vector3d()
+      if fp.translation:
+        origin = Point3d()
+        
+        trl = fp.translation.lower()
+        pos = trl.find(':') # Potential location for a parent actor
+        if pos>0:
+          parentPMO = self.getFigure().findActor(fp.translation[pos+1:], withIndex=False)
+          if not parentPMO:
+            return fp.status(C_FAIL, 'Parent [{:s}] not found for [{:s}] translation', fp.translation[pos+1:], self.getName())
+
+          origin = parentPMO.getOrigin()
+        else:
+          parentPMO = None
+        
+        if trl[0]=='(' and trl[-1]==')': # Center is given by a Tuple of float
+          tr = Point3d(eval(trl))
+          
+        elif trl.startswith('face='): # Can be Face=Material name
+          centerFaceVx = fp.geomGroup.getFaceVertexBy(fp.translation[5:])
+          tr = Point3d.barycentre(centerFaceVx).sub(bc)
+        
+        # Manage symLL[:PartName]
+        elif trl.startswith('symxy'):
+          tr = Point3d( 0.0, 0.0, -2.0*(bc.z - origin.z))
+        elif trl.startswith('symyz'):
+          tr = Point3d(-2.0*(bc.x - origin.x), 0.0, 0.0)
+        elif trl.startswith('symzx'):
+          tr = Point3d( 0.0, -2.0*(bc.y - origin.y), 0.0)
+        else:
+          return fp.status(C_FAIL,'Translation definition error for [{:s}]:{:s}', self.getName(), fp.translation) 
+        
+      PoserToken.setInitVect(lstChan, PoserToken.TRANS, tr)
+        
+      # Manage Rotation
+      vri = Vector3d()
+      if fp.rotation:
+        if fp.rotation[0]=='(' and fp.rotation[-1]==')': # Center is given by a Tuple of float
+          vri = Vector3d(eval(fp.rotation))
+        else:
+          return fp.status(C_FAIL, 'Rotation definition error for [{:s}]:{:s}', self.getName(), fp.rotation)
+        
+      PoserToken.setInitVect(lstChan, PoserToken.ROT, vri)
+        
+      # Create the list of channels
+      for gtdef in lstChan:    
+        vp = GenericTransform.createFrom(gtdef)
+        self.addAttribut(vp)
+  
+    except ValueError:
+      ret = fp.status(C_FAIL, 'Definition error for [{:s}]', self.getName())
+
+    return ret
+
 
   def CreateShaderNodeP(self, channelName, isHidden=False):
     snp = self.updateOrCreateVP(channelName, minVal=0.0, maxVal=1.0, applyLimits=True, isHidden=isHidden)
@@ -444,6 +588,9 @@ class PoserMeshedObject(StructuredAttribut):
   def setDisplayOrigin(self, b):
     self._displayOrigin = b
 
+  def setDisplayMode(self, d:'str'):
+    self._displayMode = d
+
   def getDisplayMode(self):
     return self._displayMode
 
@@ -525,65 +672,108 @@ class PoserMeshedObject(StructuredAttribut):
   # a good value.
   # @return C_OK when result is correct. C_ERROR when parameters are not correct.
   def createDeltas(self, poserRootDir, refObj, setTargetMorph, ropt):
-    ret = C_OK
-    lstCurGeom = [ self.getBaseGeomCustom(poserRootDir) ]
+    ''' Create for the current Actor the best deltas for the list of morph targets.
+    This implementation starts with an extracted refFigure for optimisation purpose.
 
+    Parameters
+    ----------
+    poserRootDir : str
+      Root Dir of the installation (to find the .OBJ files)
+
+    refFigure : str
+       name of the figure that contains initial deltas
+
+    setTargetMorph : set or list of str
+      Morphs' names to create
+      
+    ropt : ReportOption
+      Tuple of options for the report algorithm
+
+    Returns
+    -------
+    int
+      C_OK : Success
+      C_ERROR : Error in parameters
+      C_FILE_NOTFOUND : Some WaveFront files were not found
+    '''
+    ret = C_OK
+    
+    if not setTargetMorph:
+      return C_ERROR
+
+    lstCurGeom = [ self.getBaseGeomCustom(poserRootDir), ]
+    
     if isinstance(refObj, PoserMeshedObject):
       # Extract and Sort GeomCustom to get single numbered groups
-      lstRefGeom = [ refObj.getBaseGeomCustom(poserRootDir) ]
+      lstRefGeom = [ refObj.getBaseGeomCustom(poserRootDir), ]
     else: # refObj should be a Figure instance
       refPoserObject = refObj.getPoserFile()
       lstRefGeom = refPoserObject.extractAll(poserRootDir, refObj.getBodyIndex(), None)
 
-    if (lstCurGeom == None) or (lstRefGeom == None) or (setTargetMorph == None):
-      return C_ERROR
+    if not min( [ gc.isValid() for gc in lstCurGeom+lstRefGeom ] ):
+      return C_FILE_NOT_FOUND
 
     for srcGC in lstCurGeom:
       srcWG = srcGC.getWaveGeom()
 
       tabMapping = calcMapping_KDTree(srcWG, lstRefGeom, ropt.translation, ropt.maxDist)
       
-      #channelsAttr = self.getChannels()
+      self.findReportDeltas(refObj, srcGC, srcWG, tabMapping, setTargetMorph, ropt)
 
-      # Foreach morph target ==> Exploit tabMapping for srcGC
-      for channelName in setTargetMorph:
-        logging.info("Creating deltas for %s.%s", srcGC.getName(), channelName)
-        setFoundDeltas = findNewDelta(tabMapping, refObj, channelName)
-        setNewDeltas = { }
-
-        # Local Enhancement
-        Deltas.enhancement(srcWG, setFoundDeltas, setNewDeltas, ropt.enhance, ropt.boundingType, ropt.useVicinityLoop, ropt.alpha)
-        if len(setNewDeltas) > 0:
-          # Filter the result list to remove too short vectors
-          lstFinalDeltas = filterLength(setNewDeltas, ropt.minVectLen)
-
-          self.updateOrCreate(channelName, None, srcWG, lstFinalDeltas, 0)
-          logging.info("Deltas created for %s.%s : %d", srcGC.getName(), channelName, len(lstFinalDeltas))
-        else:
-          logging.warn("No Delta created for %s.%s", srcGC.getName(), channelName )
     return ret
 
+
+  def findReportDeltas(self, refObj, srcGC, srcWG, tabMapping, setTargetMorph, ropt, checkBODYParm=False):
+    ''' Optimization function. (Would be protected in Java)
+    '''
+    # Foreach morph target ==> Exploit tabMapping for srcGC
+    for channelName in setTargetMorph:
+      logging.info("Creating deltas for %s.%s", srcGC.getName(), channelName)
+      setFoundDeltas = findNewDelta(tabMapping, refObj, channelName)
+      setNewDeltas = { }
+
+      # Local Enhancement
+      Deltas.enhancement(srcWG, setFoundDeltas, setNewDeltas, ropt.enhance, ropt.boundingType, ropt.useVicinityLoop, ropt.alpha)
+      if len(setNewDeltas) > 0:
+        # Filter the result list to remove too short vectors
+        lstFinalDeltas = filterLength(setNewDeltas, ropt.minVectLen)
+
+        self.updateOrCreate(channelName, None, srcWG, lstFinalDeltas, 0)
+
+        if checkBODYParm:
+          self.getFigure().updateOrCreateValueParm(channelName)
+          
+        logging.info("Deltas created for %s.%s : %d", srcGC.getName(), channelName, len(lstFinalDeltas))
+      else:
+        logging.warn("No Delta created for %s.%s", srcGC.getName(), channelName )
+
+
+
   # Create in the current meshed object the best deltas for the alternate geoms for the list of morph targets.
   # @param poserRootDir    To find the .OBJ files
   # @param refMeshedObj    Figure that contains initial deltas
   # @param ropt            Report options
   # @param setTargetMorph  List of morphs to create
+  # @param otherRefObjs     A list of PoserMeshedObject where to find the morphs.
+  #                        Useful when the alternate geometry has been created with some fusions
   # @return    C_OK when result is correct. C_ERROR when parameters are not correct.
-  def createAlternateDeltas(self, poserRootDir, setTargetMorph, ropt):
+  def createAlternateDeltas(self, poserRootDir, setTargetMorph, ropt, otherRefObjs=None):
     ret = C_OK
+    
+    if not setTargetMorph:
+      return C_ERROR
+
     curPoserObject = self.getPoserFile()
     refMeshedObj = self
 
     # Find the actor      
     srcActor = curPoserObject.findActor(self.getName())
-    #channelsAttr = srcActor.getChannels()
 
     # Extract and load alternate geoms
     lstAltG = self.getAltGeomList()
     if len(lstAltG) > 0:
       lstCurGeom = [ ]
       # Table of alternate geometry usage valueParm indicators
-      # GenericTransform altGeomUsageIndic[] = new GenericTransform[lstAltG.size() + 1];
 
       # Find the geomChan Channel (should always exists if the list of alt geom is not empty)
       geomChan = self.getAltGeomChannel() # channelsAttr.findAttribut(PoserToken.E_geomChan)
@@ -595,9 +785,12 @@ class PoserMeshedObject(StructuredAttribut):
         lstCurGeom.append(gc)
 
       lstRefGeom = [ refMeshedObj.getBaseGeomCustom(poserRootDir), ]
+      if otherRefObjs:        
+        lstRefGeom += [ ro.getBaseGeomCustom(poserRootDir) for ro in otherRefObjs ]
 
-      if (lstCurGeom == None) or (lstRefGeom == None) or (setTargetMorph == None):
-        return C_ERROR
+      # Check File error
+      if not min( [ not gc or gc.isValid() for gc in lstCurGeom+lstRefGeom ] ):
+        return C_FILE_NOT_FOUND
 
       altno = 1
       for srcGC in lstCurGeom:
@@ -611,6 +804,13 @@ class PoserMeshedObject(StructuredAttribut):
           altChannelName = channelName + "_Alt" + str(altno)
           logging.info("Creating deltas for " + srcGC.getName() + "." + channelName + " into:" + altChannelName)
           setFoundDeltas = findNewDelta(tabMapping, refMeshedObj, channelName)
+          
+          # Search accros other Reference actors
+          if otherRefObjs:
+            for ro in otherRefObjs:
+              sfd = findNewDelta(tabMapping, ro, channelName)
+              setFoundDeltas.update(sfd)
+            
           setNewDeltas = { }
 
           # Local Enhancement
@@ -639,6 +839,118 @@ class PoserMeshedObject(StructuredAttribut):
 
     return ret
 
+  # TODO: Add logging info
+  def createMorph(self, poserRootDir, morphGeomGroup, targetMorphName, masterMorphName=None, altGeomNo=0, minVectLen=0.0):
+    '''Create a morph for the current PoserMeshedObject, based on an alternate geom.
+    The morph applies to the base geometry or to an alternate one.
+    Geometries shall have the same number of vertex (indexes matters)
+    
+    Parameters
+    ----------
+      poserRootDir : str
+        Path to the poser data installation. (Usually c:\...\Poser9) 
+      morphGeomGroup : GeomGroup
+        GeomGroup containing the morphed (translated) vertex
+      targetMorphName : str
+        Name of the local Dial (channel) to activate the morph at actor/prop level
+      masterMorphName : str, optional, default None
+        When set, defines the name of a master dial (channel) to drive the local one.
+        if it does not exist at Root (BODY) level, it will be created
+      altGeomNo : int, optional, default0
+        Index of the alternate geometry for the morph. If null (0), the default
+        geometry of the actor/prop is chosen
+      minVectLen : float, optional, default 0.0
+        Minimal length of morph vectors. None if null (0.0)
+    Returns
+    -------
+    int
+      C_OK : Success
+      C_FAIL : Bad index for the alternate geometry
+      C_BAD_DELTAINDEX : Number of vertex differ between actor/prop and geomgroup
+      C_FILE_NOTFOUND : Some Wavefront file not found
+    '''
+    ret = C_OK
+    curPoserObject = self.getPoserFile()
+    
+    # Find the actor      
+    srcActor = curPoserObject.findActor(self.getName())
+
+    # Extract and load alternate geoms
+    if altGeomNo>0:
+      lstAltG = self.getAltGeomList()
+      if altGeomNo>len(lstAltG):
+        return C_FAIL
+      
+      # Find the geomChan Channel (should always exists if the list of alt geom is not empty)
+      geomChan = self.getAltGeomChannel() # channelsAttr.findAttribut(PoserToken.E_geomChan)
+      
+      rfn = lstAltG[altGeomNo-1].getGeomFileName()
+      realpath = getRealPath(poserRootDir, rfn)
+      refGeom = GeomCustom(realpath)
+      if not refGeom or not refGeom.isValid():
+        logging.warning('Alternate Geometry file not found:%s for %s', realpath, self.getName())
+        return C_FILE_NOT_FOUND
+      
+      refWG = refGeom.getWaveGeom()
+      refGeomGrp = refWG.getGroups()[0]
+    else: # Use base geom
+      geomChan = None
+      refGeom = self.getBaseGeomCustom(poserRootDir)
+      if not refGeom or not refGeom.isValid():
+        logging.warning('Geometry not found for:%s', self.getName())
+        return C_FILE_NOT_FOUND
+      refWG = refGeom.getWaveGeom()
+
+      # Need an extract and sort to get the rigth indexes
+      refGeomGrp = refWG.extractSortGeom(refWG.groups[0].getName())
+      
+    if len(refGeomGrp.coordList)!=len(morphGeomGroup.coordList):
+      return C_BAD_DELTAINDEX
+
+    logging.info("Creating morph for %s.%s", self.getName(), targetMorphName)
+
+    # Create the dictionnary of non null deltas
+    minsq = minVectLen*minVectLen
+    setNewDeltas={}
+    for nopt,vx in enumerate(refGeomGrp.coordList):
+      movedVx = morphGeomGroup.coordList[nopt]
+      dp = DeltaPoint(nopt, movedVx.x - vx.x, movedVx.y - vx.y, movedVx.z - vx.z)
+      if dp.norme2()>minsq:
+        setNewDeltas[dp.noPt] = dp
+
+    # Create or Update the TargetGeom
+    
+    # Link to selected alternate geom if any
+    if geomChan:      
+      destGT = self.updateOrCreateVP(targetMorphName, minVal=0.0, maxVal=1.0, applyLimits=False)
+
+      targetGT = self.updateOrCreate(targetMorphName + "_Alt" + str(altGeomNo), self.getName(), refGeomGrp, setNewDeltas, srcActor.getFigure().getBodyIndex(), False)
+      
+      vod = ValueOpDelta(PoserToken.E_valueOpKey, srcActor.getFigure().getRefName(), srcActor.getName(), geomChan.getName(), 0.0)
+
+      for ia in range(1, len(lstAltG)+1):
+        vod.addValueKey(ia, 1.0 if (ia == altGeomNo) else 0.0)
+                       
+      targetGT.addVOD(vod)
+      
+      # Multiply by the value of the master channel
+      vodMul = ValueOpDelta(PoserToken.E_valueOpTimes, srcActor.getFigure().getRefName(), srcActor.getName(), targetMorphName, 0.0)
+      targetGT.addVOD(vodMul)
+
+      targetGT.setHidden(True)  
+    else:
+      destGT = self.updateOrCreate(targetMorphName, self.getName(), refGeomGrp, setNewDeltas, srcActor.getFigure().getBodyIndex(), False)    
+    
+    # If a masterMorphName is given: Create it at BODY level then link the destGT to it
+    if masterMorphName:
+      bodyAct = self.getFigure().getRootActor()
+      bodyAct.updateOrCreateVP(masterMorphName, minVal=0.0, maxVal=1.0, applyLimits=False)
+      vod = ValueOpDelta(PoserToken.E_valueOpDeltaAdd, srcActor.getFigure().getRefName(), bodyAct.getName(), masterMorphName, 1.0)
+      destGT.addVOD(vod)
+    
+    logging.info("Morph created for %s.%s : %d", refGeomGrp.getName(), targetMorphName, len(setNewDeltas))
+
+    return ret
     
   # 
   # Return the delta for a given channel and a given vertex no.
@@ -650,7 +962,7 @@ class PoserMeshedObject(StructuredAttribut):
   # @return Delta Object (Aka 3D vector)
   def findDelta(self, channelName, refNo):
     gt = self.getTargetGeom(channelName)
-    return next((dploc for dploc in gt.getDeltas().getDeltaTab() if dploc.noPt==refNo), None) if gt else None
+    return gt.getDeltas().deltaSet.get(refNo) if gt and gt.getDeltas() else None
           
 
   def deleteGTReference(self, grp, gt):
@@ -667,13 +979,27 @@ class PoserMeshedObject(StructuredAttribut):
           #  Recursive, because a group may contain another group
           self.deleteGTReference(po, gt)
 
-  # 
-  # Delete the given channel.
-  # 
-  # @param gt
   def deleteChannel(self, gt):
+    '''Delete the given channel.
+    
+    Parameter:
+    ----------
+    gt : str or PoserGenericTransform
+      The channel to delete.    
+    Returns
+    -------
+    int : C_OK when work done, C_FAIL if string input not found
+
+    '''
+    if isinstance(gt, str):
+      # Find the channel
+      gt = self._channels.get(gt.lower())
+      if not gt:
+        return C_FAIL
+      
     del self._channels[gt.getName().lower()]
     self.deleteGTReference(self._groups, gt)
+    return C_OK
 
   # 
   # Compute the first level list of channels that point to the given channel argument.
@@ -707,10 +1033,9 @@ class PoserMeshedObject(StructuredAttribut):
       gc = self.getBaseGeomCustom(poserRootDir)
       if gc and (gc.isValid()):
         dlt = gt.getDeltas()
-        dpt = dlt.getDeltaTab() if dlt else [ ]
         vertexSize = gc.getWaveGeom().getCoordListLength()
         if vertexSize == gt.getNumbDeltas():
-          for dp in dpt:
+          for dp in dlt.deltaSet.values():
             if dp.noPt >= vertexSize:
               return C_BAD_DELTAINDEX
         else:
@@ -1042,6 +1367,7 @@ class PoserActor(PoserMeshedObject):
   def __init__(self, Name = None):
     super(PoserActor, self).__init__(Name)
     self.idx = 0
+    self.setPoserType(PoserToken.E_actor)
 
   #    * @see pftk.poser.kern.PoserObject#setName(java.lang.String)
   def setName(self, n):
