@@ -95,19 +95,31 @@ class OffSA(SimpleAttribut):
 # Just read the entire line
 #  
 class TextureMapSA(SimpleAttribut):
-  def __init__(self):
-    super(TextureMapSA, self).__init__()
-    self._path = ""
-
+  def __init__(self, t=None, initPath=None):
+    super(TextureMapSA, self).__init__(n=t)
+    if initPath:
+      if isinstance(initPath, TextureMapSA):
+        self.setPoserType(initPath.getPoserType())
+        self.setPath(initPath.getPath())
+      else:
+        self.setPath("" if initPath==PoserConst.C_NO_MAP else initPath)
+    else:
+      self._path = ''
+      
   def read(self, st, val):
     self._path = val
     if self._path==PoserConst.C_NO_MAP:
       self._value = self._path
     else:
-      code, val = st.getFullLine()
+      _, val = st.getFullLine()
       self._value = self._path + "\n\t" + val
 
   def getPath(self): return self._path
+  
+  def setPath(self, p):
+    self._path = p
+    self._value = self._path + ('\n\t 0 0' if p else '')
+    
 
 #  
 # Dimension attribut.
@@ -207,7 +219,8 @@ class ValueOpDelta(SimpleAttribut):
                PoserToken.E_valueOpPlus  if c0=='+' else \
                PoserToken.E_valueOpMinus if c0=='-' else \
                PoserToken.E_valueOpTimes if c0=='*' else \
-               PoserToken.E_valueOpDivideBy
+               PoserToken.E_valueOpDivideBy if c0=='/' else \
+               PoserToken.BADTOKEN
   
       self.setPoserType(opType)
       self.setName(opType.token)
@@ -215,10 +228,12 @@ class ValueOpDelta(SimpleAttribut):
       if (opType==PoserToken.E_valueOpPlus) or (opType==PoserToken.E_valueOpMinus) or \
          (opType==PoserToken.E_valueOpTimes) or (opType==PoserToken.E_valueOpDivideBy):
         qualifiedName = expr[1:]
-      else: # case(PoserToken.E_valueOpDeltaAdd):
+      elif opType==PoserToken.E_valueOpDeltaAdd: # case(PoserToken.E_valueOpDeltaAdd):
         posstar = expr.find('*')
         self.controlRatio = float(expr[0:posstar])
         qualifiedName = expr[posstar + 1:]
+      else: # Error
+        raise ParsingErrorException('Incorrect Expression:'+expr)
   
       ptind = qualifiedName.find('.')
       if ptind < 0:
@@ -240,10 +255,10 @@ class ValueOpDelta(SimpleAttribut):
       self.l0 = src.l0
       self.l1 = src.l1
       self.l2 = src.l2
-      self.controlRatio = src.ctrlRatio
+      self.controlRatio = src.controlRatio
       
-      if src.keys:
-        for t in src.keys:
+      if src.lstValKeys:
+        for t in src.lstValKeys:
           self.addValueKey(t[0], t[1])
 
     else: # Default Creator called 
