@@ -15,8 +15,10 @@ from langutil import C_OK, C_FAIL, C_ERROR
 from pypos3d.wftk.WFBasic import Vector3d, Point3d, BoundingBox, BoundingSphere, calcMLS
 from pypos3d.wftk.Repere import Repere
 from pypos3d.wftk.PoserFileParser import PoserFileParser, ParsingErrorException
+from pypos3d.wftk.WaveGeom import WFMat, WFColor
 from pypos3d.pftk.PoserBasic import PoserConst, PoserObject, PToken, PoserToken, Lang, TAS, index, create, WBoolLine, RemoveQuotes, nodeNameNo
-from pypos3d.pftk.SimpleAttribut import KSA, ValueOpDelta, OffSA
+from pypos3d.pftk.SimpleAttribut import KSA, ValueOpDelta, OffSA, TextureMapSA,\
+  SimpleAttribut
 
  
 PtMapping = namedtuple('PtMapping', ['srcNo', 'refNo', 'refGeom', 'dist' ])
@@ -102,14 +104,14 @@ class StructuredAttribut(PoserObject):
       #  20080417 : Fix - Shall return null if not found
       return None
 
-
-  def replaceAttribut(self, attrTypeName, nattr):
-    for i,po in enumerate(self._lstAttr):
-      if po.getPoserType()==attrTypeName:
-        self._lstAttr[i] = nattr
-        return True
-
-    return False
+# Useless as we always use less 'free' attributs
+#   def replaceAttribut(self, attrTypeName, nattr):
+#     for i,po in enumerate(self._lstAttr):
+#       if po.getPoserType()==attrTypeName:
+#         self._lstAttr[i] = nattr
+#         return True
+# 
+#     return False
 
   def deleteAttribut(self, attrName):
     for po in self._lstAttr:
@@ -125,6 +127,9 @@ class StructuredAttribut(PoserObject):
   def setDirect(self, tokenID, val):
     #  Dummy (debug) implementation
     logging.warning("Unexpected Direct:" + tokenID.token + " (" + (val if val else 'null') + ") in " + str(self.__class__) + " for " + self.getPoserType().token)
+    # Workaround: Create a simple atttribut and inster it 'as usual'
+    sa = SimpleAttribut(tokenID, val)
+    self._lstAttr.append(sa)
 
   def findAttribut(self, attrTypeName:PoserToken, name=None) -> PoserObject :
     if name:
@@ -244,19 +249,32 @@ class TextureFile(StructuredAttribut):
 # This class represents the "nodeInput" element of Poser files.
 #  
 class NodeInput(StructuredAttribut):
-  def __init__(self, nodeName='', colorStr='1 1 1'):
+  def __init__(self, nodeName='', colorStr='1 1 1', src=None):
     super(NodeInput, self).__init__()
     self._poserType = PoserToken.E_nodeInput
-    self._nodeName = nodeName
-    if nodeName:
-      self._name = nodeName
-      
-    self._value    = colorStr #  22 0 32768
-    self._parmR    = "" #  NO_PARM
-    self._parmG    = "" #  NO_PARM
-    self._parmB    = "" #  NO_PARM
-    self._node     = "" #  NO_NODE
-    self._file     = "" #  "NO_MAP" 
+    
+    if src:
+      self._nodeName = src._nodeName
+      if nodeName:
+        self._name = nodeName
+        
+      self._value    = src._value
+      self._parmR    = src._parmR
+      self._parmG    = src._parmG
+      self._parmB    = src._parmB
+      self._node     = src._node
+      self._file     = src._file 
+    else:
+      self._nodeName = nodeName
+      if nodeName:
+        self._name = nodeName
+        
+      self._value    = colorStr #  22 0 32768
+      self._parmR    = "" #  NO_PARM
+      self._parmG    = "" #  NO_PARM
+      self._parmB    = "" #  NO_PARM
+      self._node     = "" #  NO_NODE
+      self._file     = "" #  "NO_MAP" 
     
   # Set the name.
   def setName(self, s):
@@ -341,17 +359,36 @@ class NodeInput(StructuredAttribut):
 #  Shader Tree Node
 #
 class Node(StructuredAttribut):
-  def __init__(self):
+  def __init__(self, src=None):
     super(Node, self).__init__()
     self._poserType = PoserToken.E_node
     self.ambiguious = True
-    self._nodeName  = ''
-    self._typeName  = ''
-    self._pos       = '0 0'
+    if src:
+      self._nodeName  = src._nodeName
+      self._typeName  = src._typeName
+      self._pos       = src._pos
+      self._collapsed = src._collapsed
+      self._showPreview = src._showPreview
+      lsrc = src.getLstAttr()
+      if lsrc:
+        for ni in lsrc:
+          nci = NodeInput(src=ni)
+          self.addAttribut(nci)
+          
+    else:
+      self._nodeName  = ''
+      self._typeName  = ''
+      self._pos       = '0 0'
+      self._collapsed = False
+      self._showPreview = False
 
   def setDirect(self, tokenID, val):
     if tokenID == PoserToken.E_name:
       self._nodeName = RemoveQuotes(val)
+    elif tokenID==PoserToken.E_inputsCollapsed:
+      self._collapsed = (val[0] == "1")
+    elif tokenID==PoserToken.E_showPreview:
+      self._showPreview = (val[0] == "1")
     else:
       logging.warning("Unexpected Direct in Node:%s (%s)",tokenID.token, val)
   
@@ -370,7 +407,12 @@ class Node(StructuredAttribut):
     else:
       self._name = s
       
-      
+  def setCollapsed(self, inputsCollapsed):
+    self._collapsed = inputsCollapsed
+    
+  def setShowPreview(self, showPreview):
+    self._showPreview = showPreview
+
   # @see deyme.v3d.poser.PoserObject#write(java.io.FileWriter)
   def write(self, fw, pfx):
     fw.write(pfx + self.getPoserType().token + ' "' + self._typeName + '" "' + self.getName() + '"\n')
@@ -379,10 +421,16 @@ class Node(StructuredAttribut):
       fw.write(nextPfx + "{\n")
       fw.write(nextPfx + 'name "' + self._nodeName + '"\n')
       fw.write(nextPfx + "pos " + self._pos + '\n')
+
+      if self._collapsed:
+        WBoolLine(fw, nextPfx, "inputsCollapsed", self._collapsed)
+      if self._showPreview:
+        WBoolLine(fw, nextPfx, "showPreview", self._showPreview)
+        
       for po in self._lstAttr:
         po.write(fw, nextPfx)
-      if len(self._lstAttr)==0:
-        fw.write('\n')
+#       if len(self._lstAttr)==0:
+#         fw.write('\n')
       fw.write(nextPfx + "}\n")
 
   def getInputByInternalName(self, inpName='Blending'):
@@ -419,6 +467,8 @@ class ShaderTree(StructuredAttribut):
   #				{
   #				name "Blender"
   #				pos 396 54
+  #       inputsCollapsed 1
+  #       showPreview 1
   #				nodeInput "Input_1"
   #					{
   #					name "Input_1"
@@ -449,7 +499,7 @@ class ShaderTree(StructuredAttribut):
   #					node NO_NODE
   #					file NO_MAP
   #					}  
-  def CreateNode(self, poserkNodeTypeCode, pos=None): # kNodeTypeCodeBLENDER or kNodeTypeCodeIMAGEMAP
+  def CreateNode(self, poserkNodeTypeCode, pos=None, inputsCollapsed=False, showPreview=False): # kNodeTypeCodeBLENDER or kNodeTypeCodeIMAGEMAP
     n = Node()
     # Generated a name according to the type
     baseName = 'Blender_' if poserkNodeTypeCode==PoserConst.kNodeTypeCodeBLENDER else 'Image_Map_'
@@ -478,34 +528,211 @@ class ShaderTree(StructuredAttribut):
 
     if pos:
       n.setLocation(pos[0], pos[1])
+    
+    n.setCollapsed(inputsCollapsed)
+    n.setShowPreview(showPreview)
     return n
  
   def AttachTreeNodes(self, nodeDest, inputName, nodeRef):
     ni = nodeDest.getInputByInternalName(inpName=inputName)
     ni._node = nodeRef.getName()
 
+
 #  
 # material class.  
 #  
 class PoserMaterial(StructuredAttribut):
   # 
   # Create the default "skin" material.
-  #    
-  def __init__(self):
+  #
+  def __init__(self, src=None):
     super(PoserMaterial, self).__init__()
+    
+    # Create the Minimal Direct attributs
+    self._KdColor = WFColor(0.992157, 0.992157, 0.992157, 1.) 
+    self._KaColor = WFColor(0., 0., 0., 1.) 
+    self._KsColor = WFColor(0.733333, 0.733333, 0.733333, 1.) 
+    self._TextureColor = WFColor(1.0, 1.0, 1.0, 1.0) 
+    self._NsExponent = 10 
+    self._tMin = 0.
+    self._tMax  = 0. 
+    self._tExpo = 0.6 
+    self._bumpStrength = 1.
+    self._ksIgnoreTexture = 0 
+    self._reflectThruLights = 1 
+    self._reflectThruKd = 0. 
+    self._textureMap = TextureMapSA(PoserToken.E_textureMap) 
+    self._bumpMap = TextureMapSA(PoserToken.E_bumpMap)   # NO_MAP
+    self._reflectionMap = TextureMapSA(PoserToken.E_reflectionMap)  # NO_MAP
+    self._transparencyMap = TextureMapSA(PoserToken.E_transparencyMap)   # NO_MAP
+    self._ReflectionColor = WFColor(0.733333, 0.733333, 0.733333, 1.) 
+    self._reflectionStrength = 1. 
+    self._shaderTree = None
+    
+    if src:
+      self.setPoserType(PoserToken.E_material)
+          
+      if isinstance(src, WFMat):
+        # Create a Material from a simple WaveFront material
+        self.setName(src.name)
+        self._KdColor = WFColor(src.kd[0], src.kd[1], src.kd[2], src.kd[3])
+        self._KaColor = WFColor(src.ka[0], src.ka[1], src.ka[2], src.ka[3])
+        self._KsColor = WFColor(src.ks[0], src.ks[1], src.ks[2], src.ks[3])
+        
+        self._textureMap = TextureMapSA(PoserToken.E_textureMap, initPath=src.map_kd)
+        
+      elif isinstance(src, PoserMaterial):
+        
+        self.setName(src.getName())
+        self._KdColor =  WFColor(src._KdColor.r, src._KdColor.g, src._KdColor.b, src._KdColor.t)
+        self._KaColor =  WFColor(src._KaColor.r, src._KaColor.g, src._KaColor.b, src._KaColor.t)
+        self._KsColor =  WFColor(src._KsColor.r, src._KsColor.g, src._KsColor.b, src._KsColor.t)
+        self._TextureColor =  WFColor(src._TextureColor.r, src._TextureColor.g, src._TextureColor.b, src._TextureColor.t)
+        self._NsExponent =  src._NsExponent 
+        self._tMin = src._tMin 
+        self._tMax  =  src._tMax  
+        self._tExpo =  src._tExpo 
+        self._bumpStrength = src._bumpStrength 
+        self._ksIgnoreTexture =  src._ksIgnoreTexture 
+        self._reflectThruLights =  src._reflectThruLights 
+        self._reflectThruKd =  src._reflectThruKd 
+        self._textureMap = TextureMapSA(initPath=src._textureMap) 
+        self._bumpMap = TextureMapSA(initPath=src._bumpMap)
+        self._reflectionMap = TextureMapSA(initPath=src._reflectionMap)
+        self._transparencyMap = TextureMapSA(initPath=src._transparencyMap)
+        self._ReflectionColor = WFColor(src._ReflectionColor.r, src._ReflectionColor.g, src._ReflectionColor.b, src._ReflectionColor.t)
+        self._reflectionStrength = src._reflectionStrength
+        if src._shaderTree:
+          self._shaderTree = ShaderTree()
+          
+          for n in src._shaderTree.getLstAttr():
+            # It is supposed that ShaderTree only contains Nodes
+            nn = Node(src=n)
+            self._shaderTree.addAttribut(nn)
+          
+          
+      else: # Illegal source
+        raise TypeError() 
 
-  def getTextureMap(self): return (self.findAttribut(PoserToken.E_textureMap)).getPath()
+  def getTextureMap(self): return self._textureMap.getPath() #if self._textureMap else None 
 
-  def getBumpMap(self): return (self.findAttribut(PoserToken.E_bumpMap)).getPath()
+  def getBumpMap(self): return self._bumpMap.getPath() #if self._bumpMap else None
 
-  def getReflectionMap(self): return (self.findAttribut(PoserToken.E_reflectionMap)).getPath()
+  def getReflectionMap(self): return self._reflectionMap.getPath() #if self._reflectionMap else None
 
-  def getTransparencyMap(self): return (self.findAttribut(PoserToken.E_transparencyMap)).getPath()
+  def getTransparencyMap(self): return self._transparencyMap.getPath() #if self._transparencyMap else None
+
+  def getShaderTree(self): return self._shaderTree
 
   def getLstNodes(self):
-    st = self.findAttribut(PoserToken.E_shaderTree)
+    st = self._shaderTree
     return [ po for po in st.getLstAttr() if isinstance(po, Node) ] if st else [ ]
 
+  def setDirect(self, tokenID, val):
+    if tokenID == PoserToken.E_KdColor:
+      tv = val.split()
+      self._KdColor = WFColor(float(tv[0]), float(tv[1]), float(tv[2]), float(tv[3]))
+      
+    elif tokenID == PoserToken.E_KaColor:
+      tv = val.split()
+      self._KaColor = WFColor(float(tv[0]), float(tv[1]), float(tv[2]), float(tv[3]))
+      
+    elif tokenID == PoserToken.E_KsColor:
+      tv = val.split()
+      self._KsColor = WFColor(float(tv[0]), float(tv[1]), float(tv[2]), float(tv[3]))
+       
+    elif tokenID == PoserToken.E_TextureColor:
+      tv = val.split()
+      self._TextureColor = WFColor(float(tv[0]), float(tv[1]), float(tv[2]), float(tv[3]))
+      
+    elif tokenID == PoserToken.E_NsExponent:
+      self._NsExponent = float(val)
+      
+    elif tokenID == PoserToken.E_tMin:
+      self._tMin = float(val)
+      
+    elif tokenID == PoserToken.E_tMax:
+      self._tMax  = float(val)
+      
+    elif tokenID == PoserToken.E_tExpo:
+      self._tExpo = float(val)
+      
+    elif tokenID == PoserToken.E_bumpStrength:
+      self._bumpStrength = float(val)
+      
+    elif tokenID == PoserToken.E_ksIgnoreTexture:
+      self._ksIgnoreTexture = int(val)
+      
+    elif tokenID == PoserToken.E_reflectThruLights:
+      self._reflectThruLights = int(val)
+      
+    elif tokenID == PoserToken.E_reflectThruKd:
+      self._reflectThruKd = float(val)
+      
+    elif tokenID == PoserToken.E_ReflectionColor:
+      tv = val.split()
+      self._ReflectionColor = WFColor(float(tv[0]), float(tv[1]), float(tv[2]), float(tv[3]))
+       
+    elif tokenID == PoserToken.E_reflectionStrength:
+      self._reflectionStrength = float(val)
+
+    else:
+      logging.warning("Unexpected Direct in Material:%s (%s)",tokenID.token, val)
+  
+  def addAttribut(self, po):
+    tokenID = po.getPoserType()
+    if tokenID == PoserToken.E_textureMap:
+      self._textureMap = po
+       
+    elif tokenID == PoserToken.E_bumpMap:
+      self._bumpMap = po
+      
+    elif tokenID == PoserToken.E_reflectionMap:
+      self._reflectionMap = po
+      
+    elif tokenID == PoserToken.E_transparencyMap:
+      self._transparencyMap = po
+    
+    elif tokenID==PoserToken.E_shaderTree:
+      self._shaderTree = po
+    else:
+      super(PoserMaterial, self).addAttribut(po)
+
+  # @see deyme.v3d.poser.PoserObject#write(java.io.FileWriter)
+  def write(self, fw, pfx):
+    fw.write(pfx + self.getPoserType().token + ' ' + self.getName() + '\n')
+    nextPfx = pfx + "\t"
+    fw.write(nextPfx + "{\n")
+    fw.write(nextPfx + 'KdColor {:f} {:f} {:f} {:f}\n'.format(self._KdColor.r, self._KdColor.g, self._KdColor.b, self._KdColor.t)) 
+    fw.write(nextPfx + 'KaColor {:f} {:f} {:f} {:f}\n'.format(self._KaColor.r, self._KaColor.g, self._KaColor.b, self._KaColor.t)) 
+    fw.write(nextPfx + 'KsColor {:f} {:f} {:f} {:f}\n'.format(self._KsColor.r, self._KsColor.g, self._KsColor.b, self._KsColor.t)) 
+    fw.write(nextPfx + 'TextureColor {:g} {:g} {:g} {:g}\n'.format(self._TextureColor.r, self._TextureColor.g, self._TextureColor.b, self._TextureColor.t)) 
+
+    fw.write(nextPfx + 'NsExponent {:f}\n'.format(self._NsExponent)) 
+    fw.write(nextPfx + 'tMin {:f}\n'.format(self._tMin)) 
+    fw.write(nextPfx + 'tMax {:f}\n'.format(self._tMax)) 
+    fw.write(nextPfx + 'tExpo {:f}\n'.format(self._tExpo)) 
+    fw.write(nextPfx + 'bumpStrength {:f}\n'.format(self._bumpStrength)) 
+    fw.write(nextPfx + 'ksIgnoreTexture {:d}\n'.format(self._ksIgnoreTexture)) 
+    fw.write(nextPfx + 'reflectThruLights {:d}\n'.format(self._reflectThruLights)) 
+    fw.write(nextPfx + 'reflectThruKd {:f}\n'.format(self._reflectThruKd)) 
+    
+    self._textureMap.write(fw, nextPfx)
+    self._bumpMap.write(fw, nextPfx)
+    self._reflectionMap.write(fw, nextPfx)
+    self._transparencyMap.write(fw, nextPfx)
+
+    fw.write(nextPfx + 'ReflectionColor {:f} {:f} {:f} {:f}\n'.format(self._ReflectionColor.r, self._ReflectionColor.g, self._ReflectionColor.b, self._ReflectionColor.t)) 
+    fw.write(nextPfx + 'reflectionStrength {:f}\n'.format(self._reflectionStrength)) 
+
+    if self._shaderTree:
+      self._shaderTree.write(fw, nextPfx)
+
+    if self._lstAttr:
+      for po in self._lstAttr:
+        po.write(fw, nextPfx)
+
+    fw.write(nextPfx + "}\n")
 
 
 
@@ -590,13 +817,13 @@ class Keys(StructuredAttribut):
             curK = KSA(int(tv[0]), float(tv[1]))
             self._hshKeys[curK._noFrame] = curK
             # Read 3 lines after (if any)
-            c0, csl, cval = st.getLine()
+            c0, _, cval = st.getLine()
             
             if c0==PoserFileParser.TT_RIGHTBRACKET: 
               break
             
-            c1, cty, *_ = st.getLine()
-            c2, ccnx, *_ = st.getLine()
+            _, cty, *_ = st.getLine()
+            _, ccnx, *_ = st.getLine()
 
             # elif vc==PoserToken.E_sl:
             curK.setSl(cval[0]=='1')
@@ -1076,6 +1303,33 @@ class GenericTransform(StructuredAttribut):
     self._transfName = src._transfName
 
 
+  @classmethod
+  def createFrom(cls, gtdef:'list from PoserToken'):
+    ''' Create a GenericTransform and initialize it with the data
+    of tables PoserToken.MINIMAL_***_CHANNEL
+    
+    Package Reserved method
+    '''
+    vp = GenericTransform(gtdef[0], gtdef[0].defaultName)
+    
+    vp.setPrintName(gtdef[0].token)
+    vp.setInterpStyleLocked(0)
+    vp.setMin(gtdef[2])
+    vp.setMax(gtdef[3])
+    vp.setInitValue(gtdef[4])      
+    vp.setForceLimits(0)
+    vp.setTrackingScale(gtdef[8])
+    
+    if gtdef[1]: #If Static channel
+      vp.setStaticValue(gtdef[4])
+      vp.setHidden(True)
+      
+    if (gtdef[0] in PoserToken.SCALES) or not gtdef[1]:
+      # Init Scales at 1.0 (100%) and other with their initvalue
+      vp.addKeyFrame(0, gtdef[4])
+      
+    return vp
+    
 
   def addKeyFrame(self, frameno, v):
     if not self._keys:
@@ -1633,6 +1887,11 @@ class SphereZone(StructuredAttribut):
   def __init__(self, Name=None):
     super(SphereZone, self).__init__()
 
+class CapsuleZone(StructuredAttribut):
+  def __init__(self, Name=None):
+    super(CapsuleZone, self).__init__(Name)
+
+
 # E_Data Tuple Key/Value for Customer Data Storage
 custData = namedtuple('custData', ['val', 'withPose', 'withMat'])
 
@@ -1766,17 +2025,19 @@ def findNewDelta(tabMapping, refObj, channelName):
             pass
     else: # refObj is supposed to be a PoserMeshedObject
       gt = refObj.getTargetGeom(channelName)
-  
-      cachDelta = gt.getDeltas().deltaSet if gt.getDeltas() else emptySet
 
-      for pm in tabMapping:
-        if pm.refGeom:
-          try:
-            srcDp = cachDelta[pm.refNo]
-            dp = DeltaPoint(pm.srcNo, srcDp.x, srcDp.y, srcDp.z)
-            setNewDeltas[dp.noPt] = dp
-          except KeyError:
-            pass
+      # FIX: More robust
+      if gt:
+        cachDelta = gt.getDeltas().deltaSet if gt.getDeltas() else emptySet
+  
+        for pm in tabMapping:
+          if pm.refGeom:
+            try:
+              srcDp = cachDelta[pm.refNo]
+              dp = DeltaPoint(pm.srcNo, srcDp.x, srcDp.y, srcDp.z)
+              setNewDeltas[dp.noPt] = dp
+            except KeyError:
+              pass
 
     return setNewDeltas
 
