@@ -3,7 +3,7 @@
 import sys, copy
 import logging
 import math
-
+import collections
 import numpy as np
 from scipy import spatial
 
@@ -142,10 +142,45 @@ class GeomGroup(object):
     lastIdx = self.getFaceLastIdx(noface)
     
     if restab:
-      restab[:] = [ Point3d(self.coordList[self.vertIdx[i]]) for i in range(startIdx, lastIdx) ] if copyPoint else [ self.coordList[self.vertIdx[i]] for i in range(startIdx, lastIdx) ]
+      restab[:] = [ Point3d(self.coordList[j]) for j in self.vertIdx[startIdx:lastIdx] ] if copyPoint else [ self.coordList[j] for j in self.vertIdx[startIdx:lastIdx] ]
       return restab
 
-    return [ Point3d(self.coordList[self.vertIdx[i]]) for i in range(startIdx, lastIdx) ] if copyPoint else [ self.coordList[self.vertIdx[i]] for i in range(startIdx, lastIdx) ]
+    return [ Point3d(self.coordList[j]) for j in self.vertIdx[startIdx:lastIdx] ] if copyPoint else [ self.coordList[j] for j in self.vertIdx[startIdx:lastIdx] ]
+
+
+  def getFaceVertexBy(self, matName, raiseExcept=True):
+    ''' Return the list of Vertex of faces with the given material name.
+    When raiseExcept is True:
+      Raise ValueError when the material does not exist or the list is empty
+    Else return an empty list
+    
+    Parameters
+    ----------
+    matName : str
+      Material Name
+      
+    raiseExcept : bool, optional, default True
+      if true, raise a ValueError
+      else may return a None list
+
+    Returns
+    -------
+    list of Point3d
+    '''
+    try:
+      matidx = self.lstMat.index(matName)
+      lstFaceNo = [ fi for fi in range(0, self.getNbFace()) if self.matIdx[fi]==matidx ]    
+      
+      if lstFaceNo:
+        return [ self.coordList[j] for j in set(i for faceno in lstFaceNo for i in self.vertIdx[self.stripCount[faceno]:self.stripCount[faceno+1]]) ]
+      
+    except ValueError:
+      pass
+
+    if raiseExcept:
+      raise ValueError('No face for '+matName)      
+    else:
+      return None
 
   #
   #
@@ -552,7 +587,6 @@ class GeomGroup(object):
 
     
 
-#   */
 
                          
     
@@ -812,12 +846,12 @@ class GeomGroup(object):
 
   # -----------------------------------------------------------------------------
   def fusion(self, angrp):
-    ''' Fusion 'ngrp' into self group.
+    ''' Fusion 'angrp' GeomGroup into self group.
     Both groups may belong to different WaveGeom
 
     Parameters
     ----------
-    ngrp : GeomGroup
+    angrp : GeomGroup
       Group to add.
     '''
     
@@ -842,6 +876,7 @@ class GeomGroup(object):
       mapVert = [-1] * nbInit      
       nbMaxVerts = len(self.coordList)      
       nbMaxTVerts = len(self.texList)
+      nbMaxNorms = len(self.normList)
 
       if (nbMaxVerts<1000) and (nbMaxTVerts<1000):
         
@@ -851,56 +886,46 @@ class GeomGroup(object):
           self.geom.coordList[:] = ngrp.geom.getCoordList()
         else:
           logging.info("Using O(n2) method: %d/%d", nbMaxVerts, nbMaxTVerts)
-          #t0 = time.clock()
-          #nbhit = 0
           for i,vi in enumerate(ngrp.vertIdx):
             nvi = mapVert[vi]
             if nvi<0:
               nvi = IndexAdd(self.coordList, Point3d(ngrp.coordList[vi]))
               mapVert[vi] = nvi
-            #else:
-              #nbhit+=1
               
             ngrp.vertIdx[i] = nvi
           
-          #dt = time.clock() - t0
-          #print('O(n2).Vertex:{0:.1f} pt/s - hit:[{1:d}/{2:d}]'.format( float(len(ngrp.vertIdx))/float(dt), nbhit, len(ngrp.vertIdx)))
-        
         if hasTexture:
           # Case of empty dest list (append few times)
           if nbMaxTVerts==0:# TexCoord Mapping are already copied and correct
             logging.info("Direct Copy of %d Texture coords", len(ngrp.texList))
             self.geom.texList[:] = ngrp.geom.getTexList()
           else:          
-            #t0 = time.clock()
+            # Create an empty mapping table
             mapTVert = [-1] * len(ngrp.texList)
-            #nbhit, nbCommon = 0,0
+
             for i,vti in enumerate(ngrp.tvertIdx):
               nvti = mapTVert[vti]
               if nvti<0:
                 nvti = IndexAdd(self.texList, TexCoord2f(ngrp.texList[vti]))
                 mapTVert[vti] = nvti
-                #nbCommon+=1
-              #else:
-                #nbhit+=1
             
               ngrp.tvertIdx[i] = nvti
     
-            #dt = time.clock() - t0
-            #print('O(n2).Text:{0:.1f} pt/s - hit:[{1:d}/{2:d}] - Common={3:d}'.format( float(len(ngrp.tvertIdx))/float(dt), nbhit, len(ngrp.tvertIdx), nbCommon))
-
-  
         if hasNorm:          
-          mapNorm = [-1] * len(ngrp.normList)
-          for i,vti in enumerate(ngrp.normIdx):
-            nvti = mapNorm[vti]
-            if nvti<0:
-              nvti = IndexAdd(self.normList, Vector3d(ngrp.normList[vti]))
-              mapNorm[vti] = nvti
-          
-            ngrp.normIdx[i] = nvti
+          if nbMaxNorms==0:
+            logging.info("Direct Copy of %d normals", len(ngrp.normList))
+            self.geom.normList[:] = ngrp.geom.getNormList()
+          else:                         
+            mapNorm = [-1] * len(ngrp.normList)
+            for i,vti in enumerate(ngrp.normIdx):
+              nvti = mapNorm[vti]
+              if nvti<0:
+                nvti = IndexAdd(self.normList, Vector3d(ngrp.normList[vti]))
+                mapNorm[vti] = nvti
+            
+              ngrp.normIdx[i] = nvti
       
-        # TODO: Proceed with Lines
+      # TODO: Proceed with Lines
       else: # Use an octree and a PaveList2D to merge coordList(s)
         logging.info("Using Octree method: %d/%d", nbMaxVerts, nbMaxTVerts)
         
@@ -912,8 +937,6 @@ class GeomGroup(object):
         tree = spatial.KDTree(npTab, leafsize=10 if nbMaxVerts<10000 else 100)
         svect = np.zeros((1,3))
         
-        #t0 = time.clock()
-
         for i,vi in enumerate(ngrp.vertIdx):
           nvi = mapVert[vi]
           if nvi<0:
@@ -933,35 +956,20 @@ class GeomGroup(object):
           
           ngrp.vertIdx[i] = nvi
           
-        #dt = time.clock() - t0
-        #print('Octree.Vertex:{0:.1f} pt/s'.format( float(len(ngrp.vertIdx))/float(dt)))
-
         if hasTexture:
           mapTVert = [-1] * len(ngrp.texList)
-          #minx, maxx, miny,maxy = sys.float_info.max, sys.float_info.min, sys.float_info.max, sys.float_info.min
           
-          #t0 = time.clock()
-          #nbhit, nbCommon = 0,0
           pl = PaveList2D(n=32, texList=self.texList)
-          #dt = time.clock() - t0
-          #print('PaveList.Create:{0:.1f} pt/s -stat:{1:s}'.format( float(len(ngrp.tvertIdx))/float(dt), pl.statStr()))
 
           for i,vti in enumerate(ngrp.tvertIdx):
-#             dt = time.clock() - t0
-#             if (dt>5.0) and (i%100==0): print('PAveList.Text[{0:d}] : {1:.1f} pt/s'.format( i, float(i)/dt))
             nvti = mapTVert[vti]
             if nvti<0:
-              texture = TexCoord2f(ngrp.texList[nvti])
+              texture = TexCoord2f(ngrp.texList[vti]) # FIX bad source texture index
               nvti = pl.IndexAdd(texture)
               mapTVert[vti] = nvti
-            #else:
-              #nbhit+=1
               
             ngrp.tvertIdx[i] = nvti
           
-          #dt = time.clock() - t0
-          #print('Octree.PaveList:{0:.1f} pt/s - hit:[{1:d}/{2:d}] - Common={3:d}'.format( float(len(ngrp.tvertIdx))/float(dt), nbhit, len(ngrp.tvertIdx), nbCommon))
-          
       # Map Material indexes
       mapMatIdx = [ IndexAdd(self.lstMat, matName) for matName in ngrp.lstMat ]
       ngrp.matIdx = [ mapMatIdx[mi] for mi in ngrp.matIdx ]
@@ -996,7 +1004,7 @@ class GeomGroup(object):
       self.tvertIdx += angrp.tvertIdx
       
     if destHasNorm:
-      self.normIdx += angrp.vertIdx
+      self.normIdx += angrp.normIdx
 
     logging.info("End Result [%s:%d faces, %d Vx, %d Tx]", self.getName(), self.getNbFace(), len(self.coordList),\
                  len(self.texList) if destHasTexture else 0)
@@ -1235,6 +1243,10 @@ class GeomGroup(object):
     
     return CoordSyst(center, eu, ev)
   
+  def calcBarycentre(self):
+    '''Compute and return the iso barycentre of the used vertex. '''
+    # Consider each vertex once
+    return Point3d.barycentre([ self.coordList[i] for i in set(self.vertIdx) ])
 
   # ---------------------------------------------------------------------------
   # TODO: Move it in WFBasic for a more general usage and wall it the vertex
@@ -1372,6 +1384,174 @@ class GeomGroup(object):
       p.inLin33(MT).add(center) 
     
     return self
+  
+  
+
+  
+  def unTriangularize(self, maxsin=FEPSILON):
+    ''' Untriangularize the current group (inplace).
+    Triangular faces are merged when:
+    * They have the same material
+    * They have a common edge
+    * The sinus of their normals (angle) is smaller than 'maxsin'
+    
+    Parameters
+    ----------
+    maxsin: float, Optional, default = FEPSILON
+      Maximal value for the sinus of the angle of both
+      normals to permit a face merge
+      
+    Return
+    ------
+    int : 
+      >=0    : Number of faces in the result group
+      C_FAIL : No face to merge
+    
+    '''
+    lstTri = []
+    nVertIdx = [] 
+    nTvertIdx = [] 
+    nNormIdx = [] 
+    nMatIdx = []
+    nStrip = [ 0, ]
+    hasTexture = self.tvertIdx
+    hasNormal = self.normIdx
+    initNbFace = self.getNbFace()
+    for noface in range(0, initNbFace):
+      startIdx = self.getFaceStartIdx(noface)
+      lastIdx = self.getFaceLastIdx(noface)
+      
+      if lastIdx-startIdx==3: # It's a triangle
+        
+        p0,p1,p2 = self.vertIdx[startIdx],self.vertIdx[startIdx+1], self.vertIdx[startIdx+2]           
+        vn = Point3d.triangle_normal(self.coordList[p0], self.coordList[p1], self.coordList[p2])
+        edges =  ( (p0,p1), (p1,p2), (p2,p0) )
+        revegdes = ( (p0,p2), (p2,p1), (p1,p0) )
+        
+        tf = TriFace(noface, p0, p1, p2, vn, edges, revegdes)
+        
+        lstTri.append(tf)
+        
+      else: # Just copy non triangular face
+        nVertIdx += self.vertIdx[startIdx:lastIdx]
+        nStrip.append(len(nVertIdx))
+        if hasTexture:
+          nTvertIdx += self.tvertIdx[startIdx:lastIdx]
+        if hasNormal: 
+          nNormIdx += self.normIdx[startIdx:lastIdx] 
+        nMatIdx.append(self.matIdx[noface])
+    
+    dn2 = maxsin*maxsin
+    # Check triangle by 2 with a filter on the material index ... but it's an O(n2) algo
+    while lstTri:
+      # Get the first triangle
+      curTri = lstTri.pop()
+      e0 = curTri.edges[0]
+      e1 = curTri.edges[1]
+      e2 = curTri.edges[2]
+      
+      mi = self.matIdx[curTri.noface]
+      
+      # Compute the list of adjacent triangles (up to 3)
+      lstAdj = [ t for t in lstTri \
+                if (self.matIdx[t.noface]==mi) and \
+                  ( (e0 in t.revedges) or (e1 in t.revedges) or (e2 in t.revedges) ) and \
+                  (curTri.vnorm.cross(t.vnorm).norme2() <= dn2) ]
+      
+      if lstAdj: # Choose the 'best' triangle to merge
+        adjTri = lstAdj[0]
+        
+        # Find the common Edge
+        ceCurTri = commEdge(curTri, adjTri)
+        lenEdge  = self.coordList[curTri.edges[ceCurTri][0]].distance( self.coordList[curTri.edges[ceCurTri][1]] )
+        
+        if len(lstAdj)>1:
+          ce = commEdge(curTri, lstAdj[1])
+          le = self.coordList[curTri.edges[ce][0]].distance( self.coordList[curTri.edges[ce][1]] )
+          if le > lenEdge:
+            adjTri = lstAdj[1]
+            ceCurTri = ce
+            
+        if len(lstAdj)>2:
+          ce = commEdge(curTri, lstAdj[2])
+          le = self.coordList[curTri.edges[ce][0]].distance( self.coordList[curTri.edges[ce][1]] )
+          if le > lenEdge:
+            adjTri = lstAdj[2]
+            ceCurTri = ce
+
+        # Remove Adjacent Triangle from the list of available faces
+        lstTri.remove(adjTri)
+          
+        # Merge curTri and adjTri
+        cea = commEdge(adjTri, curTri)
+        pp = adjTri.p2 if cea==0 else (adjTri.p0 if cea==1 else adjTri.p1)
+        if ceCurTri==0:
+          tv = [ curTri.p0, pp, curTri.p1, curTri.p2 ]
+        elif ceCurTri==1:
+          tv = [ curTri.p0, curTri.p1, pp, curTri.p2 ]
+        else:
+          tv = [ curTri.p0, curTri.p1, curTri.p2, pp ]
+
+        nVertIdx += tv
+        nStrip.append(len(nVertIdx))
+        nMatIdx.append(self.matIdx[curTri.noface])
+        
+        if hasTexture:
+          idx = self.getFaceStartIdx(adjTri.noface)
+          tp = self.tvertIdx[idx+2] if cea==0 else (self.tvertIdx[idx] if cea==1 else self.tvertIdx[idx+1])
+
+          idx = self.getFaceStartIdx(curTri.noface)          
+          if ceCurTri==0:
+            tt = [ self.tvertIdx[idx], tp, self.tvertIdx[idx+1], self.tvertIdx[idx+2] ]
+          elif ceCurTri==1:
+            tt = [ self.tvertIdx[idx], self.tvertIdx[idx+1], tp, self.tvertIdx[idx+2] ]
+          else:
+            tt = [ self.tvertIdx[idx], self.tvertIdx[idx+1], self.tvertIdx[idx+2], tp ]
+        
+          nTvertIdx += tt
+
+        if hasNormal: 
+          idx = self.getFaceStartIdx(adjTri.noface)
+          np = self.normIdx[idx+2] if cea==0 else (self.normIdx[idx] if cea==1 else self.normIdx[idx+1])
+
+          idx = self.getFaceStartIdx(curTri.noface)          
+          if ceCurTri==0:
+            tn = [ self.normIdx[idx], np, self.normIdx[idx+1], self.normIdx[idx+2] ]
+          elif ceCurTri==1:
+            tn = [ self.normIdx[idx], self.normIdx[idx+1], np, self.normIdx[idx+2] ]
+          else:
+            tn = [ self.normIdx[idx], self.normIdx[idx+1], self.normIdx[idx+2], np ]
+        
+          nNormIdx += tn
+        
+      else:
+        # The current triangle has no relevant adjacent, Add it to the new group
+        startIdx = self.getFaceStartIdx(curTri.noface)
+        lastIdx = self.getFaceLastIdx(curTri.noface)
+        nVertIdx += self.vertIdx[startIdx:lastIdx]
+        nStrip.append(len(nVertIdx))
+        nMatIdx.append(self.matIdx[curTri.noface])
+        if hasTexture:
+          nTvertIdx += self.tvertIdx[startIdx:lastIdx]
+        if hasNormal: 
+          nNormIdx += self.normIdx[startIdx:lastIdx] 
+      
+    #End of While LstTri
+    
+    # Finish the modification of the group
+    self.vertIdx[:] = nVertIdx 
+    self.stripCount[:] = nStrip
+    self.matIdx[:] = nMatIdx
+    if hasTexture:
+      self.tvertIdx[:] = nTvertIdx
+    if hasNormal: 
+      self.normIdx[:] = nNormIdx     
+    
+    return self.getNbFace() if initNbFace > self.getNbFace() else C_FAIL
+  
+TriFace = collections.namedtuple('TriFace', ['noface', 'p0', 'p1', 'p2', 'vnorm', 'edges', 'revedges'])
+def commEdge(t, adjt):
+  return 0 if (t.p0,t.p1) in adjt.revedges else (1 if (t.p1,t.p2) in adjt.revedges else (2 if (t.p2, t.p0) in adjt.revedges else -1))
 #
 
 
