@@ -105,11 +105,11 @@ def LowestIdxPos(idxTab):
       pos = i
   return pos
 
-def findCommonPoints(gBasSrcTbl, gHautSrcTbl):
-  ''' Return the sorted list of common values between input lists '''
-  r = list(set(gBasSrcTbl) & set(gHautSrcTbl))
-  r.sort()
-  return r
+# def findCommonPoints(gBasSrcTbl, gHautSrcTbl):  --> Inlined
+#   ''' Return the sorted list of common values between input lists '''
+#   r = list(set(gBasSrcTbl) & set(gHautSrcTbl))
+#   r.sort()
+#   return r
 
 # -------------------------------------------------------------------
 def IndexAdd(l,p):
@@ -448,6 +448,28 @@ class Point3d(object):
     n.normalize()
     return n
 
+  @classmethod
+  def barycentre(cls, iterPt:'iterable of Point3d'):
+    '''Compute the iso-barycenter of a set (or list) of Point3d.
+    
+    Parameters
+    ----------
+    iterPt: iterable of Point3d
+    
+    Return
+    ------
+    a Point3d with the expected center
+    None if the input iterable is None or empty    
+    '''
+    if iterPt:
+      b = Point3d()
+      for p in iterPt:
+        b.add(p)
+      b.scale(1.0/len(iterPt))
+      return b
+    else:
+      return None
+
   # End of Class Point3d ======================================================
 
 
@@ -574,35 +596,36 @@ class TexCoord2f():
 # =============================================================================#
 # Library (more or less) internal functions                                    #
 # =============================================================================#
-def __isConvexXY(v0, v1):
-  v0.normalize()
-  v1.normalize()
-  return (v0.x*v1.y - v0.y*v1.x >= 0.0)
-
-def isConvexXY(p00, p01, p10, p11):
-  s1 = __isConvexXY(Vector3d(p01.x-p00.x, p01.y-p00.y, 0.0), Vector3d(p11.x-p00.x, p11.y-p00.y, 0.0))
-  
-  return  s1==__isConvexXY(Vector3d(p10.x-p01.x, p10.y-p01.y, 0.0), Vector3d(p00.x-p01.x, p00.y-p01.y, 0.0)) and \
-    s1==__isConvexXY(Vector3d(p11.x-p10.x, p11.y-p10.y, 0.0), Vector3d(p01.x-p10.x, p01.y-p10.y, 0.0)) and \
-    s1==__isConvexXY(Vector3d(p00.x-p11.x, p00.y-p11.y, 0.0), Vector3d(p10.x-p11.x, p10.y-p11.y, 0.0)) 
-
-def isRegular(p00, p01, p10, p11):
-  v0 = Vector3d(p01.x-p00.x, p01.y-p00.y, p01.z-p00.z).normalize()
-  v1 = Vector3d(p11.x-p00.x, p11.y-p00.y, p11.z-p00.z).normalize()
-  vn = v0.cross(v1)
-  
-  if vn.norme()<FEPSILON:
-    return False
-
-  for a,b,c in [ (p10, p01, p00), ( p11, p10, p01), (p00, p11, p10) ]:
-    v0 = Vector3d(a.x-b.x, a.y-b.y, a.z-b.z).normalize()
-    v1 = Vector3d(c.x-b.x, c.y-b.y, c.z-b.z).normalize()
-    on = v0.cross(v1)
-    s = vn.dot(on)
-    if s<=0.4:
-      return False
-
-  return True
+# Code Cleaning
+# def __isConvexXY(v0, v1):
+#   v0.normalize()
+#   v1.normalize()
+#   return (v0.x*v1.y - v0.y*v1.x >= 0.0)
+# 
+# def isConvexXY(p00, p01, p10, p11):
+#   s1 = __isConvexXY(Vector3d(p01.x-p00.x, p01.y-p00.y, 0.0), Vector3d(p11.x-p00.x, p11.y-p00.y, 0.0))
+#   
+#   return  s1==__isConvexXY(Vector3d(p10.x-p01.x, p10.y-p01.y, 0.0), Vector3d(p00.x-p01.x, p00.y-p01.y, 0.0)) and \
+#     s1==__isConvexXY(Vector3d(p11.x-p10.x, p11.y-p10.y, 0.0), Vector3d(p01.x-p10.x, p01.y-p10.y, 0.0)) and \
+#     s1==__isConvexXY(Vector3d(p00.x-p11.x, p00.y-p11.y, 0.0), Vector3d(p10.x-p11.x, p10.y-p11.y, 0.0)) 
+# 
+# def isRegular(p00, p01, p10, p11):
+#   v0 = Vector3d(p01.x-p00.x, p01.y-p00.y, p01.z-p00.z).normalize()
+#   v1 = Vector3d(p11.x-p00.x, p11.y-p00.y, p11.z-p00.z).normalize()
+#   vn = v0.cross(v1)
+#   
+#   if vn.norme()<FEPSILON:
+#     return False
+# 
+#   for a,b,c in [ (p10, p01, p00), ( p11, p10, p01), (p00, p11, p10) ]:
+#     v0 = Vector3d(a.x-b.x, a.y-b.y, a.z-b.z).normalize()
+#     v1 = Vector3d(c.x-b.x, c.y-b.y, c.z-b.z).normalize()
+#     on = v0.cross(v1)
+#     s = vn.dot(on)
+#     if s<=0.4:
+#       return False
+# 
+#   return True
 
 
 def Regularity(p00, p01, p10, p11):
@@ -661,13 +684,13 @@ def Regularity(p00, p01, p10, p11):
             
   return amin, idxamin, cmin, idxcmin, vn
 
-
-def SinusXY(p00, p01, p11):
-  v0 = Vector3d(p01.x-p00.x, p01.y-p00.y, 0.0)
-  v1 = Vector3d(p11.x-p00.x, p11.y-p00.y, 0.0)
-  v0.normalize()
-  v1.normalize()
-  return v0.x*v1.y - v0.y*v1.x
+# Code Cleaning
+# def SinusXY(p00, p01, p11):
+#   v0 = Vector3d(p01.x-p00.x, p01.y-p00.y, 0.0)
+#   v1 = Vector3d(p11.x-p00.x, p11.y-p00.y, 0.0)
+#   v0.normalize()
+#   v1.normalize()
+#   return v0.x*v1.y - v0.y*v1.x
 
 # ==============================================================================
 # Edge Management
