@@ -7,39 +7,45 @@ import logging
 import traceback
 import gzip
 import bisect
+from collections import namedtuple
 
 import numpy as np
 from scipy import spatial
 
-from langutil import C_OK, C_FAIL, C_ERROR
+from langutil import C_OK, C_FAIL, C_UNKNOWN, C_ERROR, GetLowExt
 from pypos3d.wftk import WFBasic 
 from pypos3d.wftk.WFBasic import C_MISSING_FACEMAT, C_MISSING_MAT, FEPSILON, CoordSyst, FaceNormalOrder
-from pypos3d.wftk.WFBasic import LowestIdxPos, findCommonPoints, CreateLoop, FaceVisibility
+from pypos3d.wftk.WFBasic import LowestIdxPos, CreateLoop, FaceVisibility
 from pypos3d.wftk.WFBasic import TexCoord2f, Point3d, Vector3d, calcMLS, getOBJFile, Edge, IndexAdd, FaceCut, FaceSplit, CuttingData
 
 from pypos3d.wftk.PoserFileParser import OBJ_FNAT, PoserFileParser
 from pypos3d.wftk.Repere import Repere
 from pypos3d.wftk.GeomGroup import GeomGroup
 
+WFColor = namedtuple('WFColor', [ 'r', 'g', 'b', 't', 'name'])
+''' Basic 4 float color : rgb + t '''
+WFColor.__new__.__defaults__ = (0.0, 0.0, 0.0, 1.0, None) 
 
 class WFMat:
   ''' Simple WaveFront Material.
-  Only diffuse color, diffuse map and transparency are managed.
+  Only diffuse three colors (Diffuse, A, S), a diffuse map and transparency are managed.
   '''
   
-  def __init__(self, name, kd=(0.9, 0.9, 0.9), map_kd=None):
+  def __init__(self, name, kd=WFColor(0.9, 0.9, 0.9), map_kd=None):
     self.name = name
     self.d = 1.0 # Opacity
     self.kd = kd if len(kd)==4 else kd + (self.d, )
     self.map_kd = map_kd
-    # self.ka = (0.0, 0.0, 0.0)
-    # self.ks = (0.0, 0.0, 0.0)
+    self.ka = WFColor()
+    self.ks = WFColor()
     # self.ke = (0.0, 0.0, 0.0)
     # self.illum = 2.0
     
 
   @classmethod
   def readMtl(cls, libMaterialName):
+    '''Read a .mtl file and return a dictionary of WFMat.
+    '''
     logging.info("Loading Material file:%s", libMaterialName)
 
     rin = open(libMaterialName, 'rt', errors='replace') 
@@ -57,18 +63,24 @@ class WFMat:
           curMat = WFMat(rw)
           matDict[rw] = curMat
 
-        elif (lw=="ka") or (lw=="ks") or (lw=='ke') or (lw=='illum'):
-          #tv = rw.split()
-          #curMat.ka = ( float(tv[0]), float(tv[1]), float(tv[2]), float(tv[3]) )
+        elif (lw=='ke') or (lw=='illum'):
           pass
+
+        elif lw=="ka":
+          tv = rw.split()
+          curMat.ka = WFColor( float(tv[0]), float(tv[1]), float(tv[2]) )
+        
+        elif lw=="ks":
+          tv = rw.split()
+          curMat.ks = WFColor( float(tv[0]), float(tv[1]), float(tv[2]) )
         
         elif lw=="kd":
           tv = rw.split()
-          curMat.kd = (float(tv[0]), float(tv[1]), float(tv[2]), curMat.d)
+          curMat.kd = WFColor(float(tv[0]), float(tv[1]), float(tv[2]), curMat.d)
 
         elif lw=="d":
           curMat.d = float(rw)
-          curMat.kd = (curMat.kd[0], curMat.kd[1], curMat.kd[2], curMat.d)
+          curMat.kd = WFColor(curMat.kd[0], curMat.kd[1], curMat.kd[2], curMat.d)
 
         elif lw=="map_kd":
           curMat.map_kd = rw
@@ -421,47 +433,176 @@ class WaveGeom(object):
     return gret
 
 
-  # TODO Accelerate with KDTree and PaveList2D
-  def optimizeGroups(self, cleaning=False):
-    ''' Optimize the groups of this WaveGeom.
-    After Vertex and texture coordinates are uniq.
-    O(n2) long algorythm.
-    '''
-    if cleaning:
-      hasTexture = (self.texList!=None) and (len(self.texList) > 0)
-      nbInit = len(self.coordList)
-      mapVert = [-1] * nbInit
-      mapTVert = [-1] * len(self.texList) if hasTexture else None
-      
-      nCoordList = [ ]
-      nTexList = []
-      
-      for grp in self.groups:
-        for i,vi in enumerate(grp.vertIdx):
-          nvi = mapVert[vi]
-          if nvi<0:
-            nvi = IndexAdd(nCoordList, self.coordList[vi])
-            mapVert[vi] = nvi
-          
-          grp.vertIdx[i] = nvi
-        
-        if hasTexture:          
-          for i,vti in enumerate(grp.tvertIdx):
-            nvti = mapTVert[vti]
-            if nvti<0:
-              nvti = IndexAdd(nTexList, self.texList[vti])
-              mapTVert[vti] = nvti
-          
-            grp.tvertIdx[i] = nvti
 
-      self.coordList = nCoordList
-      self.texList = nTexList
+
+
+  def optimizeGroups(self, cleaning=False, radius=0.0):
+    ''' Optimize the groups of this WaveGeom.
+    Keep only Vertex, Texture coordinates and normal referenced (used) by faces.
+    if a 'cleaning' is required, it also deduplicate vertex, using cleanDupVert method.
+    
+    After Vertex and texture coordinates are uniq.
+    If cleaning is required a second optimization is performed to remove duplicated vertex.
+    
+    Parameters
+    ----------
+    cleaning: boolean, optional, default False
+      When true launch a cleanDupVert after the used vertex optimization
       
-      logging.info("Vertex Optimization=%d/%d", len(self.coordList), nbInit)
+    radius: float, optional, default 0.0
+      Maximum distance to fuse two vertexes
+    
+    Return
+    ------
+    int:
+      C_OK if the vertex table has been reduced, else C_FAIL
+    '''
+    hasTexture = (self.texList!=None) and (len(self.texList) > 0)
+    hasNormal  = (self.normList!=None) and (len(self.normList) > 0)
+    nbInit = len(self.coordList)
+    
+    # Allocate mapping tables
+    mapVert  = [-1] * nbInit
+    mapTVert = [-1] * len(self.texList) if hasTexture else None
+    mapNVert = [-1] * len(self.normList) if hasNormal else None
+    
+    for grp in self.groups:
+      for vi in grp.vertIdx:
+        # Just indicate that the vertex is used
+        mapVert[vi] = 0
+      
+      if hasTexture:          
+        for vti in grp.tvertIdx:
+          mapTVert[vti] = 0
+      
+      if hasNormal:
+        for vni in grp.normIdx:
+          mapNVert[vni] = 0
+        
+
+    # Renumber vertex and filter the initial coord list
+    nbVert = 0
+    nCoordList = [ ]
+    
+    for i,vusage in enumerate(mapVert):
+      if vusage==0:
+        mapVert[i] = nbVert
+        nCoordList.append(self.coordList[i])
+        nbVert += 1
+
+    for grp in self.groups:
+      grp.vertIdx[:] = [ mapVert[vi] for vi in grp.vertIdx ]
+    
+    self.coordList = nCoordList
+    logging.info("Unused Vertex Optimization=%d/%d", len(self.coordList), nbInit)
+    
+    # Optimize unused texture coordinates
+    if hasTexture:
+      nbInitTex = len(self.texList)
+      nTexList = []
+      nbTVert = 0
+      for i,vusage in enumerate(mapTVert):
+        if vusage==0:
+          mapTVert[i] = nbTVert
+          nTexList.append(self.texList[i])
+          nbTVert += 1
+  
+      for grp in self.groups:
+        grp.tvertIdx[:] = [ mapTVert[vti] for vti in grp.tvertIdx ]
+      self.texList = nTexList
+      logging.info("Unused Texture Optimization=%d/%d", len(self.texList), nbInitTex)
+    
+    # Optimize unused normals
+    if hasNormal:
+      nbInitNorm = len(self.normList)
+      nNormList = []
+      nbNVert = 0
+      for i,vusage in enumerate(mapNVert):
+        if vusage==0:
+          mapNVert[i] = nbNVert
+          nNormList.append(self.normList[i])
+          nbNVert += 1
+  
+      for grp in self.groups:
+        grp.normIdx[:] = [ mapNVert[vti] for vti in grp.normIdx ]
+        
+      self.normList = nNormList
+      logging.info("Unused Normal Optimization=%d/%d", len(self.normList), nbInitNorm)
     
     for gg in self.groups:
       gg.linkCoordList(self)
 
+    if cleaning:
+      self.cleanDupVert(radius)
+
+    return C_OK if nbInit>len(self.coordList) else C_FAIL
+
+  def cleanDupVert(self, radius=0.0):
+    ''' Clean duplcated vertex, ie. merge vertex that are closer than 'radius'.
+    Does not take into account whenever vertex are used or not.
+    
+    Parameters
+    ----------
+    radius: float, optional, default 0.0
+      Maximum distance to fuse two vertexes
+    
+    Return
+    ------
+    int:
+      C_OK if the vertex table has been reduced, else C_FAIL
+    '''
+    nbsrc = len(self.coordList)
+
+    logging.info("Start Cleaning for %s [%d vertex]", self.getName(), nbsrc)
+
+    # Create a numpy table Nx3
+    npTab = np.zeros( (nbsrc, 3) )
+    for refNo, p in enumerate(self.coordList):
+      npTab[refNo] = [ p.x, p.y, p.z ]
+
+    # Create an KDtree with the numpy table
+    tree = spatial.KDTree(npTab, leafsize=10 if nbsrc<10000 else 100)
+      
+    _, mapVert = tree.query(npTab, distance_upper_bound = radius if radius else np.inf)
+
+    # nCoordLst = [ p for i,p in enumerate(self.coordList) if mapVert[i]==i ]
+    nCoordLst = []
+    nMapVert = [-1] * nbsrc
+    curIdx = 0
+    for i,p in enumerate(self.coordList):
+      if mapVert[i]==i:
+        nCoordLst.append(p)
+        nMapVert[i] = curIdx
+        curIdx += 1
+      elif nMapVert[mapVert[i]]==-1:
+        # FIX appeared with scipy 1.6
+        # The point has never been seen and refers
+        # a point later in the list        
+        nCoordLst.append(p)
+        nMapVert[i] = curIdx
+        nMapVert[mapVert[i]] = curIdx
+        mapVert[mapVert[i]] = i
+        curIdx += 1          
+      else:
+        nMapVert[i] = nMapVert[mapVert[i]]
+
+    nbnew = len(nCoordLst)
+    if nbnew == nbsrc:
+      logging.info("No optimization")
+      return C_FAIL
+    
+    for gg in self.groups:
+      gg.vertIdx[:] = [ nMapVert[vi] for vi in gg.vertIdx ] 
+
+    self.setCoordList(nCoordLst)
+
+    logging.info("Optimized to %d vertex", nbnew)
+    
+    return C_OK
+
+
+
+
   #
   #    * @see deyme.v3d.wf.WaveGeom#createGeomGroup(java.lang.String)
   #    
@@ -582,8 +723,9 @@ class WaveGeom(object):
         (not self.curGroup in self.groups):
       self.groups.append(self.curGroup)      
 
-    # Close Groups
-    self.optimizeGroups()
+    # Relink Groups
+    for gg in self.groups:
+      gg.linkCoordList(self)
 
     # return a tuple of values (for Poser usage)
     return (numbVerts, numbTVerts, numbTSets, numbElems, numbSets)
@@ -949,6 +1091,34 @@ class WaveGeom(object):
 
     return ret
 
+
+  def save(self, fn):
+    ''' Write a WaveFront file in text mode or in compressed mode.
+    Use the extension to find the right mode.
+    Set current object name to 'fn'
+    
+    Parameters
+    ----------
+    fn : str
+      Full path name
+
+    Returns
+    -------
+    int
+      C_OK write without error, C_ERROR a write error has occurred
+    '''
+    if GetLowExt(fn)== 'obj': 
+      ret = self.writeOBJ(fn)
+    else:
+      ret = self.writeOBZ(fn)
+
+    self.setName(fn)
+
+    return ret
+
+
+
+
   # Find the Egde vextex based on 'findEdges' method.
   # Group is supposed to be alone in the GeomCustom.
   #   
@@ -960,15 +1130,10 @@ class WaveGeom(object):
     ng = self.getGroup(gName)
     return ng.findEdgeCoord()
 
-  # Find the common points.
-  # 
-  # @param pBasName
-  #            Name of lower element
-  # @param pHautName
-  #            Name of high element
-  # @return a table of index of Vertex of 'this'.
-  # public int[] findJonction(String pBasName, String pHautName)
-  def findJonction(self, pBasName, pHautName):
+
+  #  public int[] extractSortJonction(String pBasName, String pHautName)
+  def extractSortJonction(self, pBasName, pHautName):
+    ''' Return the list of  ... internal stuff '''
     gBas = self.getGroup(pBasName)
     if not gBas:
       logging.warning("Groupe Bas[%s] NO faces", pBasName )
@@ -983,25 +1148,20 @@ class WaveGeom(object):
 
     logging.info("Groupe Haut[%s]: %d faces", pHautName, gHaut.getNbFace()) 
 
-    #    int[] comPtTbl = util.findCommonPoints(gBas.getVertIdxTable(), gHaut.getVertIdxTable())
-    comPtTbl = findCommonPoints(gBas.vertIdx, gHaut.vertIdx)
-    logging.info("ComPoints[]: %d points", len(comPtTbl))
+    # Find common vertex index and sort them : comCoordOrig = findCommonPoints(gBas.vertIdx, gHaut.vertIdx)
+    comCoordOrig = list(set(gBas.vertIdx) & set(gHaut.vertIdx))
+    comCoordOrig.sort()
 
-    return comPtTbl
+    logging.info("ComPoints[]: %d points", len(comCoordOrig))
 
-
-  #  public int[] extractSortJonction(String pBasName, String pHautName)
-  def extractSortJonction(self, pBasName, pHautName):
-    ''' Return the list of  ... internal stuff '''
-    comCoordOrig = self.findJonction(pBasName, pHautName)
+    
     comCoord = None
 
     if comCoordOrig:
       logging.info("Extracting Jonction : %s/%s", pBasName, pHautName)
       grp = self.getGroup(pBasName)
-
       nVertIdx = self.calcGroupVertIndex(grp)
-      comCoord = [ nVertIdx.index(comCoordOrig[no]) for no in range(0, len(comCoordOrig)) ]
+      comCoord = [ nVertIdx.index(vi) for vi in comCoordOrig ]
 
     return comCoord
 
@@ -1017,17 +1177,12 @@ class WaveGeom(object):
   #           res[i] is the index in the GeomCustom vertex table.
   #  public int[] calcGroupVertIndex(GeomGroup g)
   def calcGroupVertIndex(self, grp):
-    nVertIdx = [ ]
-
-    for faceno in range(0, grp.getNbFace()):
-      startIdx = grp.stripCount[faceno]
-      lastIdx = grp.stripCount[faceno + 1]
-      for i in range(startIdx,lastIdx):
-        try:
-          nVertIdx.index(grp.vertIdx[i])
-        except:
-          nVertIdx.append(grp.vertIdx[i])
+    #     for faceno in range(0, grp.getNbFace()):
+    #       svi |= set(grp.vertIdx[grp.stripCount[faceno]:grp.stripCount[faceno + 1]])
+    svi = set( grp.vertIdx )
 
+      
+    nVertIdx = list(svi)
     nVertIdx.sort()
     return nVertIdx
 
@@ -1151,10 +1306,10 @@ class WaveGeom(object):
   
 
   #  public int findMinDist(Point3d p, int noMax, double treshhold)
-  def findMinDist(self, p, noMax, treshhold):
+  def findMinDist(self, p, noMax=-1, treshhold=FEPSILON):
     idx = -1
     minDist = sys.float_info.max
-
+    noMax = len(self.coordList) if noMax==-1 else min(noMax, len(self.coordList))
     for i in range(0, noMax):
       pe = self.coordList[i]
       d = p.distance(pe)
@@ -1234,52 +1389,7 @@ class WaveGeom(object):
       gg.invertFaceOrder()
 
 
-  #
-  # Clean duplicate vertex
-  # Not tested, Not Pythonic
-  def cleanDupVert(self, radius):
-    res = C_OK
-    logging.info("Start Cleaning for %s", self.getName())
-    nbMaxVerts = len(self.coordList)
-    tabNewIdx = [ -1 ]*nbMaxVerts
-    i = 0
-    j = 0
-    curIdx = 0
 
-    while i < nbMaxVerts:
-      # Look for first vertex to manage
-      while (i < nbMaxVerts) and (tabNewIdx[i] != -1):
-        i+=1
-
-      if i < nbMaxVerts:
-        tabNewIdx[i] = curIdx
-        j = i+1
-        while j < nbMaxVerts:
-          cd = self.coordList[i].distanceLinf(self.coordList[j])
-          if cd < radius:
-            tabNewIdx[j] = curIdx
-          j += 1
-        curIdx += 1
-    #End of While
-
-    if curIdx == nbMaxVerts:
-      return C_FAIL
-
-    nCoordLst = [None]*curIdx
-    j = 0
-    i = 0
-
-    while i < nbMaxVerts:
-      nCoordLst[j] = self.coordList[i]
-      j+=1
-      while (i < nbMaxVerts) and (tabNewIdx[i] < j): i+=1
-
-    for gg in self.groups:
-      for i,vi in enumerate(gg.vertIdx):
-        gg.vertIdx[i] = tabNewIdx[vi]
-
-    self.setCoordList(nCoordLst)
-    return res
 
   def fusion(self, inLst, outMapLst=None):
     ''' Fusion the current WaveGeom with a list of other WaveGeoms.
@@ -1499,7 +1609,7 @@ class WaveGeom(object):
 
     try:
       self.groups.remove(gg)
-      if cleaning:
+      if cleaning and self.groups:
         self.optimizeGroups(cleaning=True)
       
     except ValueError:
@@ -1780,6 +1890,67 @@ class WaveGeom(object):
       grp.fusion(ngrp)
       self.groups.remove(ngrp)
     return res
+  
+  
+  def unTriangularize(self, lstGrp=None, maxsin=FEPSILON):
+    ''' Untriangularize the current geometry (inplace).
+    Triangular faces are merged when:
+    * They have the same material
+    * They have a common edge
+    * The sinus of their normals (angle) is smaller than 'maxsin'
+    
+    Parameters
+    ----------
+    lstGrp: list of GeomGroup or String, Optional, default = None
+      List of group to 'un-triangularize'
+      If none, all groups are treated
+      else the set of groups are treated. 
+    
+    maxsin: float, Optional, default = FEPSILON
+      Maximal value for the sinus of the angle of both
+      normal to permit a merge
+      
+    Return
+    ------
+    int : 
+      >=0 : Number of faces in the result geometry
+      <0 : Error or warning
+    
+    '''
+    initNbFace = self.getNbFace()
+    logging.info("Start with %d faces", initNbFace)
+    
+    ret = C_OK
+    
+    if lstGrp==None:
+      lgrp = self.getGroups()
+    else:
+      lgrp = [ ]
+      for gd in lstGrp:
+        if isinstance(gd, str): # Input is a group Name
+          g = self.getGroup(gd)
+          if g:
+            lgrp.append(g)
+          else:
+            logging.warning('Unknown group:%s', gd)
+            ret = min(ret, C_UNKNOWN)
+        elif isinstance(gd, GeomGroup):
+          lgrp.append(gd)
+        else:
+          logging.warning('Unknown group identification:%s', str(gd))
+          ret = min(ret, C_ERROR)
+    
+    for g in lgrp:
+      nb = g.getNbFace()
+      r = g.unTriangularize(maxsin=maxsin)
+      if r<C_OK:
+        logging.info("Untriangularization of [%s] : %d faces not changed", g.getName(), nb)
+      else:
+        logging.info("Untriangularization of [%s] : %d --> %d faces", g.getName(), nb, r)
+    
+    logging.info("End with %d faces", self.getNbFace())
+
+    return self.getNbFace() if ret>=0 else ret
 # ==============================================================================
 #  End Of WaveGeom
 # ==============================================================================
@@ -2016,7 +2187,9 @@ def PlaneCut(target, centerOrRepOrPlane, eu=None, ev=None, materialName='Section
       p.y *= radialScale
 
   nwg.coordList = rep.From(nwg.coordList)
-  nwg.optimizeGroups() # Just because we've change the coordList pointer!
+  # Just because we've change the coordList pointer!
+  for gg in nwg.groups:
+    gg.linkCoordList(nwg)
 
   cd = CuttingData(ngrp, None, rep, lstFacesIdx, nbFaces, hasTexture)
   
@@ -2203,8 +2376,9 @@ def PlaneSplit(target, centerOrRepOrPlane, eu=None, ev=None, radialLimit=0.0, ma
     lstFacesIdx, nbFaces  = None, 0
 
   nwg.coordList = rep.From(nwg.coordList)
-  logging.info("Optimizing PlaneSplit for %s", target.getName())
-  nwg.optimizeGroups() # Just because we've change the coordList pointer!
+  # Relink groups, just because we've change the coordList pointer!
+  for gg in nwg.groups:
+    gg.linkCoordList(nwg)
 
   cd = CuttingData(topGrp, botGrp, rep, lstFacesIdx, nbFaces, hasTexture)
 
@@ -2621,6 +2795,8 @@ def RadialScaleRemesh(target, centerOrRepOrPlane, eu=None, ev=None, dh=0.0, ds=0
     if reScale and tabScale:
       for noloop, loop in enumerate(Loops):
         Reps[noloop].RadialScalePoint([ nwg.coordList[idx] for idx in LoopsIdx[noloop] ], tabScale[noloop])
+      # FIX: Avoid a double scaling in case of parameter mix
+      reScale = False
 
     # Put back the top group (they belong the same WaveGeom)
     # So Vertex Indexes have not been changed
